"""Basic tensor operations implemented with the multi-backend ``keras.ops``."""

from typing import Tuple, Union

import keras
import numpy as np
from keras import ops
from scipy.ndimage import _ni_support
from scipy.ndimage._filters import _gaussian_kernel1d

from zea import log
from zea.utils import map_negative_indices


def split_seed(seed, n):
    """Split a seed into n seeds for reproducible random ops.

    Supports `keras.random.SeedGenerator <https://keras.io/api/random/#seedgenerator-class>`_
    and `JAX random keys <https://jax.readthedocs.io/en/latest/jax.random.html#jax.random.PRNGKey>`_.

    Args:
        seed: None, jax.Array, or keras.random.SeedGenerator.
        n (int): Number of seeds to generate.

    Returns:
        list: List of n seeds (JAX keys, SeedGenerator, or None).
    """
    # If seed is None, return a list of None
    if seed is None:
        return [None for _ in range(n)]

    # If seed is a JAX key, split it into n keys
    if keras.backend.backend() == "jax":
        import jax

        return jax.random.split(seed, n)

    # For other backends, we have to use Keras SeedGenerator
    else:
        assert isinstance(seed, keras.random.SeedGenerator), (
            "seed must be a SeedGenerator when not using JAX."
        )

        # Just duplicate the SeedGenerator
        return [seed for _ in range(n)]


def is_jax_prng_key(x):
    """Distinguish between jax.random.PRNGKey() and jax.random.key()"""
    if keras.backend.backend() == "jax":
        import jax

        return isinstance(x, jax.Array) and x.shape == (2,) and x.dtype == jax.numpy.uint32
    else:
        return False


def add_salt_and_pepper_noise(image, salt_prob, pepper_prob=None, seed=None):
    """Adds salt and pepper noise to the input image.

    Args:
        image (ndarray): The input image, must be of type float32 and normalized between 0 and 1.
        salt_prob (float): The probability of adding salt noise to each pixel.
        pepper_prob (float, optional): The probability of adding pepper noise to each pixel.
            If not provided, it will be set to the same value as `salt_prob`.
        seed: A Python integer or instance of
            `keras.random.SeedGenerator`.
            Used to make the behavior of the initializer
            deterministic. Note that an initializer seeded with an integer
            or None (unseeded) will produce the same random values
            across multiple calls. To get different random values
            across multiple calls, use as seed an instance
            of `keras.random.SeedGenerator`.

    Returns:
        ndarray: The noisy image with salt and pepper noise added.
    """
    if pepper_prob is None:
        pepper_prob = salt_prob

    if salt_prob == 0.0 and pepper_prob == 0.0:
        return image

    assert ops.dtype(image) == "float32", "Image should be of type float32."

    noisy_image = ops.copy(image)

    # Add salt noise
    salt_mask = keras.random.uniform(ops.shape(image), seed=seed) < salt_prob
    noisy_image = ops.where(salt_mask, 1.0, noisy_image)

    # Add pepper noise
    pepper_mask = keras.random.uniform(ops.shape(image), seed=seed) < pepper_prob
    noisy_image = ops.where(pepper_mask, 0.0, noisy_image)

    return noisy_image


def extend_n_dims(arr, axis, n_dims):
    """Extend the number of dimensions of an array.

    Inserts 'n_dims' ones at the specified axis.

    Args:
        arr: The input array.
        axis: The axis at which to insert the new dimensions.
        n_dims: The number of dimensions to insert.

    Returns:
        The array with the extended number of dimensions.

    Raises:
        AssertionError: If the axis is out of range.
    """
    assert axis <= ops.ndim(arr), (
        "Axis must be less than or equal to the number of dimensions in the array"
    )
    assert axis >= -ops.ndim(arr) - 1, (
        "Axis must be greater than or equal to the negative number of dimensions minus 1"
    )
    axis = ops.ndim(arr) + axis + 1 if axis < 0 else axis

    # Get the current shape of the array
    shape = ops.shape(arr)

    # Create the new shape, inserting 'n_dims' ones at the specified axis
    new_shape = shape[:axis] + (1,) * n_dims + shape[axis:]

    # Reshape the array to the new shape
    return ops.reshape(arr, new_shape)


def func_with_one_batch_dim(
    func,
    tensor,
    n_batch_dims: int,
    batch_size: int | None = None,
    func_axis: int | None = None,
    **kwargs,
):
    """Wraps a function to apply it to an input tensor with one or more batch dimensions.

    The function will be executed in parallel on all batch elements.

    Args:
        func (function): The function to apply to the image.
            Will take the `func_axis` output from the function.
        tensor (Tensor): The input tensor.
        n_batch_dims (int): The number of batch dimensions in the input tensor.
            Expects the input to start with n_batch_dims batch dimensions. Defaults to 2.
        batch_size (int, optional): Integer specifying the size of the batch for
            each step to execute in parallel. Defaults to None, in which case the function
            will run everything in parallel.
        func_axis (int, optional): If `func` returns mulitple outputs, this axis will be returned.
        **kwargs: Additional keyword arguments to pass to the function.

    Returns:
        The output tensor with the same batch dimensions as the input tensor.

    Raises:
        ValueError: If the number of batch dimensions is greater than the rank of the input tensor.
    """
    # Extract the shape of the batch dimensions from the input tensor
    batch_dims = ops.shape(tensor)[:n_batch_dims]

    # Extract the shape of the remaining (non-batch) dimensions
    other_dims = ops.shape(tensor)[n_batch_dims:]

    # Reshape the input tensor to merge all batch dimensions into one
    reshaped_input = ops.reshape(tensor, [-1, *other_dims])

    # Apply the given function to the reshaped input tensor
    if batch_size is None:
        reshaped_output = func(reshaped_input, **kwargs)
    else:
        reshaped_output = batched_map(func, reshaped_input, batch_size=batch_size)

    # If the function returns multiple outputs, select the one corresponding to `func_axis`
    if isinstance(reshaped_output, (tuple, list)):
        if func_axis is None:
            raise ValueError(
                "func_axis must be specified when the function returns multiple outputs."
            )
        reshaped_output = reshaped_output[func_axis]

    # Extract the shape of the output tensor after applying the function (excluding the batch dim)
    output_other_dims = ops.shape(reshaped_output)[1:]

    # Reshape the output tensor to restore the original batch dimensions
    return ops.reshape(reshaped_output, [*batch_dims, *output_other_dims])


def matrix_power(matrix, power):
    """Compute the power of a square matrix.

    Should match the
    [numpy](https://numpy.org/doc/stable/reference/generated/numpy.linalg.matrix_power.html)
    implementation.

    Parameters:
        matrix (array-like): A square matrix to be raised to a power.
        power (int): The exponent to which the matrix is to be raised.
                    Must be a non-negative integer.
    Returns:
        array-like: The resulting matrix after raising the input matrix to the specified power.

    """
    if power == 0:
        return ops.eye(matrix.shape[0])
    if power == 1:
        return matrix
    if power % 2 == 0:
        half_power = matrix_power(matrix, power // 2)
        return ops.matmul(half_power, half_power)
    return ops.matmul(matrix, matrix_power(matrix, power - 1))


def boolean_mask(tensor, mask, size=None):
    """Apply a boolean mask to a tensor.

    Args:
        tensor (Tensor): The input tensor.
        mask (Tensor): The boolean mask to apply.
        size (int, optional): The size of the output tensor. Only used for Jax backend if you
            want to trace the function. Defaults to None.

    Returns:
        Tensor: The masked tensor.
    """
    if keras.backend.backend() == "jax" and size is not None:
        import jax.numpy as jnp

        indices = jnp.where(mask, size=size)  # Fixed size allows Jax tracing
        return tensor[indices]
    elif keras.backend.backend() == "tensorflow":
        import tensorflow as tf

        return tf.boolean_mask(tensor, mask)
    else:
        return tensor[mask]


if keras.backend.backend() == "jax":
    import jax.numpy as jnp

    def nonzero(x, size=None, fill_value=None):
        """Return the indices of the elements that are non-zero.

        Args:
            x (Tensor): Input tensor.
            size (int, optional): optional static integer specifying the number of nonzero
                entries to return. If there are more nonzero elements than the specified size,
                then indices will be truncated at the end. If there are fewer nonzero elements
                than the specified size, then indices will be padded with fill_value.
            fill_value (int, optional): Value to fill in case there are not enough
                non-zero elements. Defaults to None.
        """
        return jnp.nonzero(x, size=size, fill_value=fill_value)

else:

    def nonzero(x, size=None, fill_value=None):
        """Return the indices of the elements that are non-zero."""
        return ops.nonzero(x)


def flatten(tensor, start_dim=0, end_dim=-1):
    """Should be similar to: https://pytorch.org/docs/stable/generated/torch.flatten.html"""
    # Get the shape of the input tensor
    old_shape = ops.shape(tensor)

    # Adjust end_dim if it's negative
    end_dim = ops.ndim(tensor) + end_dim if end_dim < 0 else end_dim

    # Create a new shape with -1 in the flattened dimensions
    new_shape = [*old_shape[:start_dim], -1, *old_shape[end_dim + 1 :]]

    # Reshape the tensor
    return ops.reshape(tensor, new_shape)


def batch_cov(x, rowvar=True, bias=False, ddof=None):
    """Compute the batch covariance matrices of the input tensor.

    Args:
        x (Tensor): Input tensor of shape (..., m, n) where m is the number of features
            and n is the number of observations.
        rowvar (bool, optional): If True, each row represents a variable,
            while each column represents an observation. If False, each column represents
            a variable, while each row represents an observation. Defaults to True.
        bias (bool, optional): If True, the biased estimator of the covariance is computed.
            If False, the unbiased estimator is computed. Defaults to False.
        ddof (int, optional): Delta degrees of freedom. The divisor used in the calculation
            is (num_obs - ddof), where num_obs is the number of observations.
            If ddof is not specified, it is set to 0 if bias is True, and 1 if bias is False.
            Defaults to None.

    Returns:
        Tensor: Batch covariance matrices of shape (..., m, m) if rowvar=True,
                or (..., n, n) if rowvar=False.
    """
    # Ensure the input has at least 3 dimensions
    if ops.ndim(x) == 2:
        x = x[None]

    if not rowvar:
        x = ops.moveaxis(x, -1, -2)

    num_obs = x.shape[-1]

    if ddof is None:
        ddof = 0 if bias else 1

    # Subtract the mean from each observation
    mean_x = ops.mean(x, axis=-1, keepdims=True)
    x_centered = x - mean_x

    # Compute the covariance using einsum
    cov_matrices = ops.einsum("...ik,...jk->...ij", x_centered, x_centered) / (num_obs - ddof)
    return cov_matrices


def patched_map(f, xs, patches: int, jit=True, **batch_kwargs):
    """Wrapper around `batched_map` for patching.

    Allows you to specify the number of patches rather than the batch size.
    """
    assert patches > 0, "Number of patches must be greater than 0."

    if patches == 1:
        return f(xs)
    else:
        length = ops.shape(xs)[0]
        batch_size = np.ceil(length / patches).astype(int)
        return batched_map(f, xs, batch_size, jit, **batch_kwargs)


def batched_map(f, xs, batch_size=None, jit=True, **batch_kwargs):
    """Map a function over leading array axes.

    Args:
        f (callable): Function to apply element-wise over the first axis.
        xs (Tensor): Values over which to map along the leading axis.
        batch_size (int, optional): Size of the batch for each step. Defaults to None,
            in which case the function will be equivalent to `ops.map`, and thus map over
            the leading axis.
        jit (bool, optional): If True, use a jitted version of the function for
            faster batched mapping. Else, loop over the data with the original function.
        batch_kwargs (dict, optional): Additional keyword arguments (tensors) to
            batch along with xs. Must have the same first dimension size as xs.

    Returns:
        The mapped tensor(s).

    Idea taken from: https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.map.html
    """
    if batch_kwargs is None:
        batch_kwargs = {}

    # Ensure all batch kwargs have the same leading dimension as xs.
    if batch_kwargs:
        assert all(
            ops.shape(xs)[0] == ops.shape(v)[0] for v in batch_kwargs.values() if v is not None
        ), "All batch kwargs must have the same first dimension size as xs."

    total = ops.shape(xs)[0]
    # TODO: could be rewritten with ops.cond such that it also works for jit=True.
    if not jit and batch_size is not None and total <= batch_size:
        return f(xs, **batch_kwargs)

    ## Non-jitted version: simply iterate over batches.
    if not jit:
        bs = batch_size or 1  # Default batch size to 1 if not specified.
        outputs = []
        for i in range(0, total, bs):
            idx = slice(i, i + bs)
            current_kwargs = {k: v[idx] for k, v in batch_kwargs.items()}
            outputs.append(f(xs[idx], **current_kwargs))
        return ops.concatenate(outputs, axis=0)

    ## Jitted version.

    # Helper to create the batched function for use with ops.map.
    def create_batched_f(kw_keys):
        def batched_f(inputs):
            x, *kw_values = inputs
            kw = dict(zip(kw_keys, kw_values))
            return f(x, **kw)

        return batched_f

    if batch_size is None:
        batched_f = create_batched_f(list(batch_kwargs.keys()))
        return ops.map(batched_f, (xs, *batch_kwargs.values()))

    # Pad and reshape primary tensor.
    xs_padded = pad_array_to_divisible(xs, batch_size, axis=0)
    new_shape = (-1, batch_size) + ops.shape(xs_padded)[1:]
    xs_reshaped = ops.reshape(xs_padded, new_shape)

    # Pad and reshape batch_kwargs similarly.
    reshaped_kwargs = {}
    for k, v in batch_kwargs.items():
        if v is None:
            reshaped_kwargs[k] = None
        else:
            v_padded = pad_array_to_divisible(v, batch_size, axis=0)
            reshaped_kwargs[k] = ops.reshape(v_padded, (-1, batch_size) + ops.shape(v_padded)[1:])

    batched_f = create_batched_f(list(reshaped_kwargs.keys()))
    out = ops.map(batched_f, (xs_reshaped, *reshaped_kwargs.values()))
    out_reshaped = ops.reshape(out, (-1,) + ops.shape(out)[2:])
    return out_reshaped[:total]  # Remove any padding added.


if keras.backend.backend() == "jax":
    # For jit purposes
    def _get_padding(N, remainder):
        return N - remainder if remainder != 0 else 0

else:

    def _get_padding(N, remainder):
        return ops.where(remainder != 0, N - remainder, 0)


def pad_array_to_divisible(arr, N, axis=0, mode="constant", pad_value=None):
    """Pad an array to be divisible by N along the specified axis.

    Args:
        arr (Tensor): The input array to pad.
        N (int): The number to which the length of the specified axis should be divisible.
        axis (int, optional): The axis along which to pad the array. Defaults to 0.
        mode (str, optional): The padding mode to use. Defaults to 'constant'.
            One of `"constant"`, `"edge"`, `"linear_ramp"`,
            `"maximum"`, `"mean"`, `"median"`, `"minimum"`,
            `"reflect"`, `"symmetric"`, `"wrap"`, `"empty"`,
            `"circular"`. Defaults to `"constant"`.
        pad_value (float, optional): The value to use for padding when mode='constant'.
            Defaults to None. If mode is not `constant`, this value should be None.

    Returns:
        Tensor: The padded array.
    """
    # Get the length of the specified axis
    length = ops.shape(arr)[axis]

    # Calculate how much padding is needed for the specified axis
    remainder = length % N
    padding = _get_padding(N, remainder)

    # Create a tuple with (before, after) padding for each axis
    pad_width = [(0, 0)] * ops.ndim(arr)  # No padding for other axes
    pad_width[axis] = (0, padding)  # Padding for the specified axis

    # Pad the array
    padded_array = ops.pad(arr, pad_width, mode=mode, constant_values=pad_value)

    return padded_array


def interpolate_data(subsampled_data, mask, order=1, axis=-1):
    """Interpolate subsampled data along a specified axis using `map_coordinates`.

    Args:
        subsampled_data (ndarray): The data subsampled along the specified axis.
            Its shape matches `mask` except along the subsampled axis.
        mask (ndarray): Boolean array with the same shape as the full data.
            `True` where data is known.
        order (int, optional): The order of the spline interpolation. Default is `1`.
        axis (int, optional): The axis along which the data is subsampled. Default is `-1`.

    Returns:
        ndarray: The data interpolated back to the original grid.

    ValueError: If `mask` does not indicate any missing data or if `mask` has `False`
        values along multiple axes.
    """
    mask = ops.cast(mask, "bool")
    # Check that mask indicates subsampled data along the specified axis
    if ops.sum(mask) == 0:
        raise ValueError("Mask does not indicate any known data.")

    if ops.sum(mask) == ops.prod(mask.shape):
        raise ValueError("Mask does not indicate any missing data.")

    # make sure subsampled data corresponds with number of 1s in the mask
    assert len(ops.where(mask)[0]) == ops.prod(subsampled_data.shape), (
        "Subsampled data does not match the number of 1s in the mask."
    )

    assert subsampled_data.ndim == 1, "Subsampled data should be a flattened 1D array"

    assert mask.ndim == 2, "Currently only 2D interpolation supported"

    # Get the indices of the known data points
    known_indices = ops.stack(ops.where(mask), axis=-1)

    # Get the indices of the unknown data points
    unknown_indices = ops.stack(ops.where(~mask), axis=-1)

    # map the unknown indices to the new coordinate system
    # which basically is range(0, mask.shape[axis]) for each axis
    # but with the gaps removed
    interp_coords = []
    subsampled_shape = []
    axis = axis if axis >= 0 else mask.ndim + axis

    for _axis in range(mask.ndim):
        length_axis = mask.shape[_axis]
        if _axis == axis:
            indices = ops.where(
                ops.any(~mask, axis=tuple(i for i in range(mask.ndim) if i != _axis))
            )[0]
            # unknown indices
            indices = map_indices_for_interpolation(indices)
            subsampled_shape.append(length_axis - len(indices))
        else:
            # broadcast indices
            indices = ops.arange(length_axis, dtype="float32")
            subsampled_shape.append(length_axis)

        interp_coords.append(indices)

    # create the grid of coordinates for the interpolation
    interp_coords = ops.meshgrid(*interp_coords, indexing="ij")
    # should be of shape (mask.ndim, -1)

    subsampled_data = ops.reshape(subsampled_data, subsampled_shape)

    # Use map_coordinates to interpolate the data
    interpolated_data = ops.image.map_coordinates(
        subsampled_data,
        interp_coords,
        order=order,
    )

    interpolated_data = ops.reshape(interpolated_data, -1)

    # now distirubute the interpolated data back to the original grid
    output_data = ops.zeros_like(mask, dtype=subsampled_data.dtype)

    output_data = ops.scatter_update(output_data, unknown_indices, interpolated_data)

    # Get the values at the known data points
    known_values = ops.reshape(subsampled_data, (-1,))
    output_data = ops.scatter_update(
        output_data,
        known_indices,
        known_values,
    )

    return output_data


def is_monotonic(array, increasing=True):
    """Checks if a given 1D array is monotonic.

    Either entirely non-decreasing or non-increasing.

    Args:
        array (ndarray): A 1D numpy array.

    Returns:
        bool: True if the array is monotonic, False otherwise.
    """
    # Convert to numpy array to handle general cases
    array = ops.array(array)

    # Check if the array is non-decreasing or non-increasing
    if increasing:
        return ops.all(array[1:] >= array[:-1])
    return ops.all(array[1:] <= array[:-1])


def map_indices_for_interpolation(indices):
    """Interpolates a 1D array of indices with gaps.

    Maps a 1D array of indices with gaps to a 1D array where gaps
    would be between the integers.

    Used in the `interpolate_data` function.

    Args:
        (indices): A 1D array of indices with gaps.
    Returns:
        (indices): mapped to a 1D array where gaps would be between
        the integers

    There are two segments here of length 4 and 2

    Example:
        >>> indices = [5, 6, 7, 8, 12, 13, 19]
        >>> mapped_indices = [5, 5.25, 5.5, 5.75, 8, 8.5, 12.5]

    """
    indices = ops.array(indices, dtype="int32")

    assert is_monotonic(indices, increasing=True), "Indices should be monotonically increasing"

    gap_starts = ops.where(indices[1:] - indices[:-1] > 1)[0]
    gap_starts = ops.concatenate([ops.array([0]), gap_starts + 1], axis=0)

    gap_lengths = ops.concatenate(
        [gap_starts[1:] - gap_starts[:-1], ops.array([len(indices) - gap_starts[-1]])],
        axis=0,
    )

    cumul_gap_lengths = ops.cumsum(gap_lengths)
    cumul_gap_lengths = ops.concatenate([ops.array([0]), cumul_gap_lengths], axis=0)

    gap_start_values = ops.take(indices, gap_starts)
    mapped_starts = gap_start_values - cumul_gap_lengths[:-1]
    mapped_starts = ops.cast(mapped_starts, "float32")
    gap_lengths = ops.cast(gap_lengths, "float32")

    spacing = 1 / (gap_lengths + 1)

    # Vectorized creation of gap_length entries between the start and end
    mapped_indices = ops.concatenate(
        [
            (mapped_starts[i] + spacing[i]) + spacing[i] * ops.arange(gap_lengths[i])
            for i in range(len(gap_lengths))
        ],
        axis=0,
    )
    mapped_indices -= 1

    return mapped_indices


def stack_volume_data_along_axis(data, batch_axis: int, stack_axis: int, number: int):
    """Stacks tensor data along a specified stack axis.

    Stack tensor data along a specified stack axis by splitting it into blocks along the batch axis.

    Args:
        data (Tensor): Input tensor to be stacked.
        batch_axis (int): Axis along which to split the data into blocks.
        stack_axis (int): Axis along which to stack the blocks.
        number (int): Number of slices per stack.

    Returns:
        Tensor: Reshaped tensor with data stacked along stack_axis.

    Example:
        .. code-block:: python

            import keras

            data = keras.random.uniform((10, 20, 30))
            # stacking along 1st axis with 2 frames per block
            stacked_data = stack_volume_data_along_axis(data, 0, 1, 2)
            stacked_data.shape
    """
    blocks = int(ops.ceil(data.shape[batch_axis] / number))
    data = pad_array_to_divisible(data, axis=batch_axis, N=blocks, mode="reflect")
    data = ops.split(data, blocks, axis=batch_axis)
    data = ops.stack(data, axis=batch_axis)
    # put batch_axis in front
    data = ops.transpose(
        data,
        (
            batch_axis + 1,
            *range(batch_axis + 1),
            *range(batch_axis + 2, data.ndim),
        ),
    )
    data = ops.concatenate(list(data), axis=stack_axis)
    return data


def split_volume_data_from_axis(data, batch_axis: int, stack_axis: int, number: int, padding: int):
    """Splits previously stacked tensor data back to its original shape.

    This function reverses the operation performed by `stack_volume_data_along_axis`.

    Args:
        data (Tensor): Input tensor to be split.
        batch_axis (int): Axis along which to restore the blocks.
        stack_axis (int): Axis from which to split the stacked data.
        number (int): Number of slices per stack.
        padding (int): Amount of padding to remove from the result.

    Returns:
        Tensor: Reshaped tensor with data split back to original format.

    Example:
        .. code-block:: python

            import keras

            data = keras.random.uniform((20, 10, 30))
            split_data = split_volume_data_from_axis(data, 0, 1, 2, 2)
            split_data.shape
    """
    if data.shape[stack_axis] == 1:
        # in this case it was a broadcasted axis which does not need to be split
        return data
    data = ops.split(data, number, axis=stack_axis)
    data = ops.stack(data, axis=batch_axis + 1)
    # combine the unstacked axes (dim 1 and 2)
    total_block_size = data.shape[batch_axis] * data.shape[batch_axis + 1]
    data = ops.reshape(
        data,
        (*data.shape[:batch_axis], total_block_size, *data.shape[batch_axis + 2 :]),
    )

    # cut off padding
    if padding > 0:
        indices = ops.arange(data.shape[batch_axis] - padding + 1)
        data = ops.take(data, indices, axis=batch_axis)

    return data


def compute_required_patch_overlap(image_shape, patch_shape):
    """Compute required overlap between patches to cover the entire image.

    Args:
        image_shape: Tuple of (height, width)
        patch_shape: Tuple of (patch_height, patch_width)

    Returns:
        Tuple of (overlap_y, overlap_x)

        Or None if there is no overlap that will result in integer number of patches
        given the image and patch shapes.
    """
    assert len(image_shape) == 2, "image_shape must be a tuple of (height, width)"
    assert len(patch_shape) == 2, "patch_shape must be a tuple of (patch_height, patch_width)"

    assert all(image_shape[i] >= patch_shape[i] for i in range(2)), (
        "patch_shape must be equal or smaller than image_shape"
    )

    image_y, image_x = image_shape
    patch_y, patch_x = patch_shape

    # Calculate number of patches needed in each dimension
    n_patch_y = max(1, int(ops.ceil(image_y / patch_y)))
    n_patch_x = max(1, int(ops.ceil(image_x / patch_x)))

    # Calculate new overlap only if we have more than one patch
    new_overlap = (
        ((patch_y * n_patch_y - image_y) / (n_patch_y - 1) if n_patch_y > 1 else 0),
        ((patch_x * n_patch_x - image_x) / (n_patch_x - 1) if n_patch_x > 1 else 0),
    )

    # check if can be integer
    if not all(ops.isclose(new_overlap, ops.round(new_overlap))):
        return

    new_overlap = tuple(map(int, new_overlap))

    return new_overlap


def compute_required_patch_shape(image_shape, patch_shape, overlap):
    """Compute required patch shape to cover the entire image.

    Compute patch_shape closest to the original patch_shape that will result
    in integer number of patches given the image and overlap.

    Args:
        image_shape: Tuple of (height, width)
        patch_shape: Tuple of (patch_height, patch_width)
        overlap: Tuple of (overlap_y, overlap_x)

    Returns:
        Tuple of (patch_shape_y, patch_shape_x)

        or None if there is no patch_shape that will result in integer number of patches
        given the image and overlap.
    """
    image_y, image_x = image_shape
    overlap_y, overlap_x = overlap
    patch_y, patch_x = patch_shape

    def compute_patch_size(image_size, patch_size, overlap):
        n_patches = (image_size - overlap) // (patch_size - overlap)
        new_patch_size = (image_size + (n_patches - 1) * overlap) / n_patches
        return int(new_patch_size)

    new_patch_y = compute_patch_size(image_y, patch_y, overlap_y)
    new_patch_x = compute_patch_size(image_x, patch_x, overlap_x)

    if (image_y - new_patch_y) % (new_patch_y - overlap_y) != 0 or (image_x - new_patch_x) % (
        new_patch_x - overlap_x
    ) != 0:
        return None

    return new_patch_y, new_patch_x


def check_patches_fit(
    image_shape: tuple, patch_shape: tuple, overlap: Union[int, Tuple[int, int]]
) -> tuple:
    """Checks if patches with overlap fit an integer amount in the original image.

    Args:
        image_shape: A tuple representing the shape of the original image.
        patch_size: A tuple representing the shape of the patches.
        overlap: A float representing the overlap between patches.

    Returns:
        A tuple containing a boolean indicating if the patches fit an integer amount
        in the original image and the new image shape if the patches do not fit.

    Example:
        .. code-block:: python

            image_shape = (10, 10)
            patch_shape = (4, 4)
            overlap = (2, 2)
            patches_fit, new_shape = check_patches_fit(image_shape, patch_shape, overlap)
            patches_fit
            new_shape
    """
    if overlap:
        stride = (np.array(patch_shape) - np.array(overlap)).astype(int)
    else:
        stride = (np.array(patch_shape)).astype(int)
        overlap = (0, 0)

    stride_y, stride_x = stride
    patch_y, patch_x = patch_shape
    image_y, image_x = image_shape

    if (image_y - patch_y) % stride_y != 0 or (image_x - patch_x) % stride_x != 0:
        new_shape = (
            (image_y - patch_y) // stride_y * stride_y + patch_y,
            (image_x - patch_x) // stride_x * stride_x + patch_x,
        )
        # new_patch_shape = tuple(map(int, new_patch_shape))
        new_patch_shape = compute_required_patch_shape(image_shape, patch_shape, overlap)

        # Calculate new overlap only if we have more than one patch
        new_overlap = compute_required_patch_overlap(image_shape, patch_shape)

        msg = (
            "patches with overlap do not fit an integer amount in the original image. "
            f"Cropping image to closest dimensions that work: {new_shape}. "
        )

        if new_patch_shape is not None:
            msg += f"Alternatively, change patch shape to: {new_patch_shape} "

        if new_overlap is not None:
            msg += f"or change overlap to: {new_overlap}"

        log.warning(msg)

        return False, new_shape
    return True, image_shape


def images_to_patches(
    images: keras.KerasTensor,
    patch_shape: Union[int, Tuple[int, int]],
    overlap: Union[int, Tuple[int, int]] = None,
) -> keras.KerasTensor:
    """Creates patches from images.

    Args:
        images (Tensor): input images [batch, height, width, channels].
        patch_shape (int or tuple, optional): Height and width of patch. Defaults to 4.
        overlap (int or tuple, optional): Overlap between patches in px. Defaults to None.

    Returns:
        patches (Tensor): batch of patches of size:
            [batch, #patch_y, #patch_x, patch_size_y, patch_size_x, #channels].

    Example:
        .. code-block:: python

            import keras

            images = keras.random.uniform((2, 8, 8, 3))
            patches = images_to_patches(images, patch_shape=(4, 4), overlap=(2, 2))
            patches.shape
    """
    assert len(images.shape) == 4, (
        f"input array should have 4 dimensions, but has {len(images.shape)} dimensions"
    )
    assert isinstance(patch_shape, int) or len(patch_shape) == 2, (
        f"patch_shape should be an integer or a tuple of length 2, but is {patch_shape}"
    )
    assert isinstance(overlap, (int, type(None))) or len(overlap) == 2, (
        f"overlap should be an integer or a tuple of length 2, but is {overlap}"
    )

    batch_size, *image_shape, n_channels = images.shape

    if isinstance(patch_shape, int):
        patch_shape = (patch_shape, patch_shape)
    if isinstance(overlap, int):
        overlap = (overlap, overlap)

    patch_size_y, patch_size_x = patch_shape

    patches_fit, image_shape = check_patches_fit(image_shape, patch_shape, overlap)
    if not patches_fit:
        images = images[:, : image_shape[0], : image_shape[1], :]

    if overlap:
        stride = (np.array(patch_shape) - np.array(overlap)).astype(int)
    else:
        stride = np.array(patch_shape).astype(int)

    # assert that stride is never smaller than 0 or larger than patch_shape
    stride = np.maximum(stride, 1)
    stride = np.minimum(stride, patch_shape)
    assert np.all(stride <= patch_shape), "Stride should be smaller than patch shape"
    assert np.all(stride >= 0), "Stride should be larger than 0"

    ## create patches using ops (this operation is too memory intensive)
    # patches = ops.image.extract_patches(
    #     images, size=patch_shape, strides=list(stride), padding="valid"
    # )

    ## manual solution instead
    patches_list = []
    for i in range(0, image_shape[0] - patch_size_y + 1, stride[0]):
        row_patches = []
        for j in range(0, image_shape[1] - patch_size_x + 1, stride[1]):
            patch = images[:, i : i + patch_size_y, j : j + patch_size_x, :]
            row_patches.append(patch)
        patches_list.append(ops.stack(row_patches, axis=1))
    patches = ops.stack(patches_list, axis=1)

    _, n_patch_y, n_patch_x, *_ = patches.shape

    shape = [batch_size, n_patch_y, n_patch_x, patch_size_y, patch_size_x, n_channels]
    patches = ops.reshape(patches, shape)
    return patches


def patches_to_images(
    patches: keras.KerasTensor,
    image_shape: tuple,
    overlap: Union[int, Tuple[int, int]] = None,
    window_type="average",
) -> keras.KerasTensor:
    """Reconstructs images from patches.

    Args:
        patches (Tensor): Array with batch of patches to convert to batch of images.
            [batch_size, #patch_y, #patch_x, patch_size_y, patch_size_x, n_channels]
        image_shape (Tuple): Shape of output image. (height, width, channels)
        overlap (int or tuple, optional): Overlap between patches in px. Defaults to None.
        window_type (str, optional): Type of stitching to use. Defaults to 'average'.
            Options: 'average', 'replace'.

    Returns:
        images (Tensor): Reconstructed batch of images from batch of patches.

    Example:
        .. code-block:: python

            import keras

            patches = keras.random.uniform((2, 3, 3, 4, 4, 3))
            images = patches_to_images(patches, image_shape=(8, 8, 3), overlap=(2, 2))
            images.shape
    """
    # Input validation
    assert len(image_shape) == 3, "image_shape must have 3 dimensions: (height, width, channels)."
    assert len(patches.shape) == 6, (
        "patches must have 6 dimensions: [batch_size, n_patch_y, n_patch_x, "
        "patch_size_y, patch_size_x, n_channels]."
    )
    assert window_type in [
        "average",
        "replace",
    ], "window_type must be one of 'average', or 'replace'."

    # Extract dimensions
    batch_size, n_patches_y, n_patches_x, patch_size_y, patch_size_x, _ = patches.shape
    dtype = patches.dtype

    if isinstance(overlap, int):
        overlap = (overlap, overlap)
    if overlap is None:
        overlap = (0, 0)

    stride_y, stride_x = np.array([patch_size_y, patch_size_x]) - np.array(overlap)

    # Initialize the output tensor (image) and mask
    images = keras.ops.zeros((batch_size, *image_shape), dtype=dtype)
    mask = keras.ops.zeros((batch_size, *image_shape), dtype=dtype)

    # Loop through each patch
    for i in range(n_patches_y):
        for j in range(n_patches_x):
            start_y = i * stride_y
            start_x = j * stride_x
            patch = patches[:, i, j]

            if window_type == "replace":
                # Replace pixels directly with the current patch
                images = keras.ops.slice_update(images, [0, start_y, start_x, 0], patch)
            else:
                # Add the current patch to the image
                images = keras.ops.slice_update(
                    images,
                    [0, start_y, start_x, 0],
                    images[
                        :,
                        start_y : start_y + patch_size_y,
                        start_x : start_x + patch_size_x,
                        :,
                    ]
                    + patch,
                )
                # Update the mask for averaging
                mask = keras.ops.slice_update(
                    mask,
                    [0, start_y, start_x, 0],
                    mask[
                        :,
                        start_y : start_y + patch_size_y,
                        start_x : start_x + patch_size_x,
                        :,
                    ]
                    + 1,
                )

    if window_type == "average":
        # Normalize overlapping regions if needed
        images = keras.ops.where(mask > 0, images / mask, images)

    return images


def reshape_axis(data, newshape: tuple, axis: int):
    """Reshape data along axis.

    Args:
        data (tensor): input data.
        newshape (tuple): new shape of data along axis.
        axis (int): axis to reshape.

    Example:
        .. code-block:: python

            import keras

            data = keras.random.uniform((3, 4, 5))
            newshape = (2, 2)
            reshaped_data = reshape_axis(data, newshape, axis=1)
            reshaped_data.shape
    """
    axis = map_negative_indices([axis], data.ndim)[0]
    shape = list(ops.shape(data))  # list
    shape = shape[:axis] + list(newshape) + shape[axis + 1 :]
    return ops.reshape(data, shape)


def _gaussian_filter1d(array, kernel, radius, cval=None, axis=-1, mode="symmetric"):
    if keras.backend.backend() == "torch":
        assert mode == "constant", (
            "Only constant padding is for sure correct in torch."
            "Symmetric padding produces different results in torch compared to tensorflow..."
        )

    # Pad input along the specified axis.
    pad_width = [(0, 0)] * array.ndim
    pad_width[axis] = (radius, radius)
    padded = ops.pad(array, pad_width, mode=mode, constant_values=cval)

    # Move the convolution axis to the last axis.
    moved = ops.moveaxis(padded, axis, -1)  # shape: (..., length)
    orig_shape = moved.shape
    length = orig_shape[-1]

    # Collapse all non-convolution dimensions into the batch.
    reshaped = ops.reshape(moved, (-1, length, 1))  # shape: (batch, length, in_channels=1)

    # Reshape kernel for convolution: expected shape (kernel_size, in_channels, out_channels)
    kernel_size = kernel.shape[0]
    kernel_reshaped = ops.reshape(kernel, (kernel_size, 1, 1))

    # Run the convolution using 'VALID' padding.
    conv_result = ops.depthwise_conv(
        reshaped,
        kernel_reshaped,
        padding="valid",
        data_format="channels_last",
    )

    # Reshape the convolved result back to the padded shape.
    new_length = conv_result.shape[1]
    conv_result = ops.reshape(conv_result, (*orig_shape[:-1], new_length))

    # Move the convolution axis back to its original position.
    result = ops.moveaxis(conv_result, -1, axis)
    return result


def gaussian_filter1d(array, sigma, axis=-1, order=0, mode="symmetric", truncate=4.0, cval=None):
    """1-D Gaussian filter.

    Args:
        array (Tensor): The input array.
        sigma (float or tuple): Standard deviation for Gaussian kernel. The standard deviations
            of the Gaussian filter are given for each axis as a sequence, or as a single number,
            in which case it is equal for all axes.
        order (int or Tuple[int]): The order of the filter along each axis is given as a sequence of
            integers, or as a single number. An order of 0 corresponds to convolution with a
            Gaussian kernel. A positive order corresponds to convolution with that derivative
            of a Gaussian. Default is 0.
        mode (str, optional): Padding mode for the input image. Default is 'symmetric'.
            See [keras docs](https://www.tensorflow.org/api_docs/python/tf/keras/ops/pad) for
            all options and [tensoflow docs](https://www.tensorflow.org/api_docs/python/tf/pad)
            for some examples. Note that the naming differs from scipy.ndimage.gaussian_filter!
        cval (float, optional): Value to fill past edges of input if mode is 'constant'.
            Default is None.
        truncate (float, optional): Truncate the filter at this many standard deviations.
            Default is 4.0.
        axes (Tuple[int], optional): If None, input is filtered along all axes. Otherwise, input
            is filtered along the specified axes. When axes is specified, any tuples used for
            sigma, order, mode and/or radius must match the length of axes. The ith entry in
            any of these tuples corresponds to the ith entry in axes.
    """
    # Determine the effective kernel radius and generate the Gaussian kernel
    radius = int(round(truncate * sigma))
    kernel = _gaussian_kernel1d(sigma, order, radius).astype(
        ops.dtype(array)
    )  # shape: (kernel_size,)

    # Reverse the kernel for odd orders to mimic correlation (SciPy behavior)
    if order % 2:
        kernel = kernel[::-1]

    return _gaussian_filter1d(array, kernel, radius, cval, axis, mode)


def gaussian_filter(
    array,
    sigma,
    order: int | Tuple[int] = 0,
    mode: str = "symmetric",
    cval: float | None = None,
    truncate: float = 4.0,
    axes: Tuple[int] = None,
):
    """Multidimensional Gaussian filter.

    If you want to use this function with jax.jit, you can set:
    `static_argnames=("truncate", "sigma")`

    Args:
        array (Tensor): The input array.
        sigma (float or tuple): Standard deviation for Gaussian kernel. The standard deviations
            of the Gaussian filter are given for each axis as a sequence, or as a single number,
            in which case it is equal for all axes.
        order (int or Tuple[int]): The order of the filter along each axis is given as a sequence of
            integers, or as a single number. An order of 0 corresponds to convolution with a
            Gaussian kernel. A positive order corresponds to convolution with that derivative
            of a Gaussian. Default is 0.
        mode (str, optional): Padding mode for the input image. Default is 'symmetric'.
            See [keras docs](https://www.tensorflow.org/api_docs/python/tf/keras/ops/pad) for
            all options and [tensoflow docs](https://www.tensorflow.org/api_docs/python/tf/pad)
            for some examples. Note that the naming differs from scipy.ndimage.gaussian_filter!
        cval (float, optional): Value to fill past edges of input if mode is 'constant'.
            Default is None.
        truncate (float, optional): Truncate the filter at this many standard deviations.
            Default is 4.0.
        axes (Tuple[int], optional): If None, input is filtered along all axes. Otherwise, input
            is filtered along the specified axes. When axes is specified, any tuples used for
            sigma, order, mode and/or radius must match the length of axes. The ith entry in
            any of these tuples corresponds to the ith entry in axes.
    """
    axes = _ni_support._check_axes(axes, array.ndim)
    num_axes = len(axes)
    orders = _ni_support._normalize_sequence(order, num_axes)
    sigmas = _ni_support._normalize_sequence(sigma, num_axes)
    modes = _ni_support._normalize_sequence(mode, num_axes)
    axes = [(axes[ii], sigmas[ii], orders[ii], modes[ii]) for ii in range(num_axes)]
    if len(axes) > 0:
        for (
            axis,
            sigma,
            order,
            mode,
        ) in axes:
            output = gaussian_filter1d(array, sigma, axis, order, mode, truncate, cval)
            array = output
    else:
        output = array
    return output


def resample(x, n_samples, axis=-2, order=1):
    """Resample tensor along axis.

    Similar to scipy.signal.resample.

    Args:
        x: input tensor.
        n_samples: number of samples after resampling.
        axis: axis to resample along.
        order: interpolation order (1=linear).

    Returns:
        Resampled tensor.
    """
    shape = ops.shape(x)
    rank = len(shape)

    # Move axis-to-resample to position 1
    perm = list(range(rank))
    perm_axis1 = perm.copy()
    perm_axis1[axis], perm_axis1[1] = perm_axis1[1], perm_axis1[axis]
    x = ops.transpose(x, perm_axis1)

    # Shape after transpose
    shape = ops.shape(x)
    batch_size = shape[0]
    old_n = shape[1]
    other_dims = shape[2:]

    # Create sampling grid
    batch_coords = ops.arange(batch_size, dtype="float32")  # (batch_size,)
    new_coords = ops.linspace(0.0, ops.cast(old_n - 1, dtype="float32"), n_samples)  # (n_samples,)
    other_coords = [ops.arange(d, dtype="float32") for d in other_dims]

    # Meshgrid
    grid = ops.meshgrid(
        batch_coords, new_coords, *other_coords, indexing="ij"
    )  # list of (batch_size, n_samples, ...)
    coord_grid = ops.stack(grid, axis=0)  # shape: (rank, batch_size, n_samples, ...)

    # Interpolate
    resampled = ops.image.map_coordinates(x, coord_grid, order=order)

    # Inverse transpose to restore original axis order
    inv_perm = [perm_axis1.index(i) for i in range(rank)]
    resampled = ops.transpose(resampled, inv_perm)

    return resampled


def fori_loop(lower, upper, body_fun, init_val, disable_jit=False):
    """For loop allowing for non-jitted for loop with same signature as jax.

    Args:
        lower (int): Lower bound of the loop.
        upper (int): Upper bound of the loop.
        body_fun (function): Function to be executed in the loop.
        init_val (any): Initial value for the loop.
        disable_jit (bool, optional): If True, disables JIT compilation. Defaults to False.
    """
    if not disable_jit:
        return ops.fori_loop(lower, upper, body_fun, init_val)

    # Fallback to a non-jitted for loop
    val = init_val
    for i in range(lower, upper):
        val = body_fun(i, val)
    return val


def L2(x):
    """L2 norm of a tensor.

    Implementation of L2 norm: https://mathworld.wolfram.com/L2-Norm.html
    """
    return ops.sqrt(ops.sum(x**2))


def linear_sum_assignment(cost):
    """Greedy linear sum assignment.

    Args:
        cost (Tensor): Cost matrix of shape (n, n).
    Returns:
        Tuple: Row indices and column indices for assignment.

    Returns row indices and column indices for assignment.
    """
    n = ops.shape(cost)[0]
    assigned_true = ops.zeros((n,), dtype="bool")
    row_ind = []
    col_ind = []
    for i in range(n):
        mask = 1.0 - ops.cast(assigned_true, "float32")
        masked_cost = cost[i] + (1.0 - mask) * 1e6
        idx = int(ops.argmin(masked_cost))
        row_ind.append(i)
        col_ind.append(idx)
        assigned_true = keras.ops.scatter_update(assigned_true, [[idx]], [True])
    return np.array(row_ind), np.array(col_ind)


def sinc(x, eps=keras.config.epsilon()):
    """Sinc function."""
    return ops.sin(x + eps) / (x + eps)


if keras.backend.backend() == "tensorflow":

    def safe_vectorize(
        pyfunc,
        excluded=None,
        signature=None,
    ):
        """Just a wrapper around ops.vectorize.

        Because tensorflow does not support multiple arguments to ops.vectorize(func)(...)
        We will just map the function manually.
        """

        def _map(*args):
            outputs = []
            for i in range(ops.shape(args[0])[0]):
                outputs.append(pyfunc(*[arg[i] for arg in args]))
            return ops.stack(outputs)

        return _map

else:

    def safe_vectorize(pyfunc, excluded=None, signature=None):
        """Just a wrapper around ops.vectorize."""
        return ops.vectorize(pyfunc, excluded=excluded, signature=signature)



"""Weights & Biases (wandb) tooling."""

from pathlib import Path

import wandb


def model_directory_from_wandb(workspace, name):
    """Get model directory from wandb name."""

    api = wandb.Api()
    runs = api.runs(workspace, filters={"display_name": name})
    if len(runs) == 0:
        raise ValueError(f"No runs found with name {name} in workspace {workspace}")
    if len(runs) > 1:
        raise ValueError(f"Multiple runs found with name {name} in workspace {workspace}")
    run = runs[0]

    if run.config.get("run_dir") is None:
        raise ValueError(f"Run {name} does not have a 'run_dir' in its config")

    return Path(run.config["run_dir"])



"""Jax utilities for zea."""

import jax
import numpy as np


def on_device_jax(func, inputs, device, return_numpy=False, **kwargs):
    """Applies a JAX function to inputs on a specified device.

    Args:
        func (callable): The function to apply.
        inputs (ndarray): Input array.
        device (str): Device string, e.g. ``'cuda'``, ``'gpu'``, or ``'cpu'``.
        return_numpy (bool, optional): Whether to convert output
            data back to numpy. Defaults to False.
        **kwargs: Additional keyword arguments to be passed to the ``func``.

    Returns:
        jax.numpy.DeviceArray or ndarray: The output data.

    Raises:
        AssertionError: If ``func`` is not a function from the JAX library.

    Note:
        This function converts the ``inputs`` array to a JAX array and moves
        it to the specified ``device``. It then applies the ``func`` function to the inputs
        and returns the output data. If the output is a dictionary, it extracts the first value
        from the dictionary. If ``return_numpy`` is True, it converts the output data back to a
        numpy array before returning.

    Example:
        .. code-block:: python

            import jax.numpy as jnp


            def square(x):
                return x**2


            inputs = [1, 2, 3, 4, 5]
            device = "gpu"
            output = on_device_jax(square, inputs, device)
    """
    device = device.split(":")
    if len(device) == 2:
        device_type, device_number = device
        device_number = int(device_number)
    else:
        # if no device number is specified, use the first device
        device_type = device[0]
        device_number = 0

    if device_number > len(jax.devices(device_type)):
        raise ValueError(
            f"Device {device} is not available from JAX devices: {jax.devices(device_type)}"
        )

    jax_device = jax.devices(device_type)[device_number]

    with jax.default_device(jax_device):
        outputs = func(inputs, **kwargs)

    if isinstance(outputs, dict):
        outputs = list(outputs.values())[0]

    if return_numpy:
        outputs = np.array(outputs)

    return outputs



"""Pressure field computation for ultrasound imaging.

This module provides routines for automatic computation of the acoustic pressure field
used for compounding multiple transmit (Tx) events in ultrasound imaging.

The pressure field is computed by simulating the acoustic response of the probe and
medium for each transmit event. The computation involves:

- Subdividing each probe element into sub-elements to satisfy the Fraunhofer approximation.
- Calculating the distances and angles between each grid point and each sub-element.
- Computing the frequency response of the probe and the pulse spectrum.
- Summing the contributions from all relevant frequencies, taking into account
  transmit delays, apodization, and directivity.
- Optionally normalizing and thresholding the resulting field for use in
  transmit compounding or adaptive beamforming.

The main entry point is :func:`compute_pfield`, which returns a normalized pressure
field array for all transmit events.

"""

import keras
import numpy as np
from keras import ops

from zea import log
from zea.internal.cache import cache_output
from zea.tensor_ops import sinc


def _abs_sinc(x):
    return sinc(ops.abs(x))


@cache_output(verbose=True)
def compute_pfield(
    sound_speed,
    center_frequency,
    bandwidth_percent,
    n_el,
    probe_geometry,
    tx_apodizations,
    grid,
    t0_delays,
    frequency_step=4,
    db_thresh=-1,
    downsample=10,
    downmix=4,
    alpha=1,
    percentile=10,
    norm=True,
    verbose=True,
):
    """Compute the pressure field for ultrasound imaging.

    Args:
        sound_speed (float): Speed of sound in the medium.
        center_frequency (float): Center frequency of the probe in Hz.
        bandwidth_percent (float): Bandwidth of the probe, pulse-echo 6dB fractional bandwidth (%)
        n_el (int): Number of elements in the probe.
        probe_geometry (array): Geometry of the probe elements.
        tx_apodizations (array): Transmit apodization values.
        grid (array): Grid points where the pressure field is computed of shape (Nz, Nx, 3).
        t0_delays (array): Transmit delays for each transmit event.
        frequency_step (int, optional): Frequency step. Default is 4.
            Higher is faster but less accurate.
        db_thresh (int, optional): dB threshold. Default is -1.
            Higher is faster but less accurate.
        downsample (int, optional): Downsample the grid for faster computation.
            Default is 10. Higher is faster but less accurate.
        downmix (int, optional): Downmixing the frequency to facilitate a smaller grid.
            Default is 4. Higher requires lower number of grid points but is less accurate.
        alpha (float, optional): Exponent to 'sharpen or smooth' the weighting. Higher is sharper.
            Only works when norm is True. Default is 1.
        percentile (int, optional): minimum percentile threshold to keep in the weighting.
            Only works when norm is True. Higher is more aggressive. Default is 10.
        norm (bool, optional): per pixel normalization (True) or unnormalized (False)
        verbose (bool, optional): Whether to print progress.

    Returns:
        ops.array: The (normalized) pressure field (across tx events) of shape (n_tx, Nz, Nx).
    """
    # medium params
    alpha_db = 0  # currently we ignore attenuation in the compounding

    # probe params
    center_frequency = center_frequency / downmix  # downmixing the frequency

    # pulse params
    num_waveforms = 1  # number of waveforms in the pulse

    # array params
    probe_geometry = ops.convert_to_tensor(probe_geometry, dtype="float32")

    pitch = probe_geometry[1, 0] - probe_geometry[0, 0]  # element pitch

    kerf = 0.1 * pitch  # for now this is hardcoded
    element_width = pitch - kerf

    num_transmits = len(tx_apodizations)

    # %------------------------------------%
    # % POINT LOCATIONS, DISTANCES & GRIDS %
    # %------------------------------------%

    # subdivide elements into sub elements or not? (to satisfy Fraunhofer approximation)
    lambda_min = sound_speed / (center_frequency * (1 + bandwidth_percent / 200))
    num_sub_elements = ops.ceil(element_width / lambda_min)

    x_orig = ops.convert_to_tensor(grid[:, :, 0], dtype="float32")
    z_orig = ops.convert_to_tensor(grid[:, :, 2], dtype="float32")

    size_orig = ops.shape(x_orig)

    # Nearest-neighbor downsampling the grid
    x = x_orig[::downsample, ::downsample]
    z = z_orig[::downsample, ::downsample]
    size_downsampled = ops.shape(x)

    # Coordinates of the points where pressure is needed
    x = ops.reshape(x, (-1,))
    z = ops.reshape(z, (-1,))

    # Centers of the transducer elements (x- and z-coordinates)
    xe = (ops.arange(0.0, n_el) - (n_el - 1) / 2) * pitch
    ze = ops.zeros(n_el)
    the = ops.zeros(n_el)

    # Centroids of the sub-elements
    seg_length = element_width / num_sub_elements
    tmp = (
        -element_width / 2
        + seg_length / 2
        + ops.arange(0, num_sub_elements, dtype=seg_length.dtype) * seg_length
    )
    xi = tmp
    zi = ops.zeros((int(num_sub_elements),))

    # Distances between the points and the transducer elements
    x_expanded = x[:, None, None]
    xi_expanded = xi[None, :, None]
    xe_expanded = xe[None, None, :]

    dxi = x_expanded - xi_expanded - xe_expanded

    z_expanded = z[:, None, None]
    zi_expanded = zi[None, :, None]
    ze_expanded = ze[None, None, :]

    d2 = dxi**2 + (z_expanded - zi_expanded - ze_expanded) ** 2
    r = ops.sqrt(d2)
    r_flat = ops.reshape(r, (-1,))

    # Angle between the normal to the transducer and the line joining
    # the point and the transducer
    eps = keras.config.epsilon()
    theta = ops.arcsin((dxi + eps) / (ops.sqrt(d2) + eps)) - the
    sin_theta = ops.sin(theta)

    pulse_width = num_waveforms / center_frequency  # temporal pulse width
    wc = 2 * np.pi * center_frequency

    def pulse_spectrum(w):
        imag = _abs_sinc(pulse_width * (w - wc) / 2) - _abs_sinc(pulse_width * (w + wc) / 2)
        return 1j * ops.cast(imag, "complex64")

    # FREQUENCY RESPONSE of the ensemble PZT + probe
    w_bandwidth = bandwidth_percent * wc / 100  # angular frequency bandwidth
    p_shape = ops.log(126) / ops.log(eps + 2 * wc / w_bandwidth)

    def probe_spectrum(w):
        # Calculate the normalized frequency difference
        freq_diff = ops.abs(w - wc)
        # Calculate the denominator for normalization
        denom = (w_bandwidth / 2) / (ops.log(2) ** (1 / p_shape))
        # Raise the normalized difference to the power of p_shape
        exponent = (freq_diff / denom) ** p_shape
        # Apply the negative sign and exponential
        return ops.exp(-exponent)

    p_list = []

    if verbose:
        log.info("Computing pressure field for all transmits")
        progbar = keras.utils.Progbar(num_transmits, unit_name="transmits")
    for j in range(0, num_transmits):
        # delays and apodization of transmit event
        delays_tx = ops.convert_to_tensor(t0_delays[j], dtype="float32")
        idx_nan = ops.isnan(delays_tx)
        delays_tx = ops.where(idx_nan, 0, delays_tx)

        tx_apodization = ops.convert_to_tensor(tx_apodizations[j])
        idx_nan = ops.isnan(tx_apodization)
        tx_apodization = ops.where(idx_nan, 0, tx_apodization)
        tx_apodization = ops.squeeze(tx_apodization)

        # The frequency response is a pulse-echo (transmit + receive) response.
        # The spectrum of the pulse (pulse_spectrum) will be then multiplied
        # by the frequency-domain tapering window of the transducer (probe_spectrum)
        # The frequency step df is chosen to avoid interferences due to
        # inadequate discretization.
        # df = frequency step (must be sufficiently small):
        # One has exp[-i(k r + w delay)] = exp[-2i pi(f r/c + f delay)] in the Eq.
        # One wants: the phase increment 2pi(df r/c + df delay) be < 2pi.
        # Therefore: df < 1/(r/c + delay).

        delays_tx_flat = ops.reshape(delays_tx, (-1,))

        df = 1 / (ops.max(r_flat / sound_speed) + ops.max(delays_tx_flat))
        df = frequency_step * df

        # FREQUENCY SAMPLES
        num_freq = 2 * ops.cast(ops.ceil(center_frequency / df), "int32") + 1
        freq = ops.linspace(0, 2 * center_frequency, num_freq)
        df = freq[1]

        # keep the significant components only by using db_thresh
        spectrum = ops.abs(
            pulse_spectrum(2 * np.pi * freq)
            * ops.cast(probe_spectrum(2 * np.pi * freq), "complex64")
        )
        gain_db = 20 * ops.log10(eps + spectrum / (ops.max(spectrum)))
        idx = gain_db > db_thresh

        freq = freq[idx]

        pulse_spect = pulse_spectrum(2 * np.pi * freq)
        probe_spect = probe_spectrum(2 * np.pi * freq)

        # Exponential arrays of size [numel(x) n_el num_sub_elements]
        kw = 2 * np.pi * freq[0] / sound_speed
        kwa = alpha_db / 8.69 * freq[0] / 1e6 * 1e2

        r_complex = ops.cast(r, dtype="complex64")
        kwa = ops.cast(kwa, dtype="complex64")
        mod_out = ops.cast(ops.mod(kw * r, 2 * np.pi), dtype="complex64")
        exp_arr = ops.exp(-kwa * r_complex + 1j * mod_out)

        # Exponential array for the increment wavenumber dk
        dkw = 2 * np.pi * df / sound_speed
        dkwa = alpha_db / 8.69 * df / 1e6 * 1e2
        dkw = ops.cast(dkw, dtype="complex64")
        dkwa = ops.cast(dkwa, dtype="complex64")

        exp_df = ops.exp((-dkwa + 1j * dkw) * r_complex)

        exp_arr = exp_arr / ops.sqrt(r_complex)
        exp_arr = exp_arr * ops.cast(ops.min(ops.sqrt(r)), "complex64")  # normalize the field

        center_wavenumber = 2 * np.pi * center_frequency / sound_speed
        directivity = _abs_sinc(center_wavenumber * seg_length / 2 * sin_theta)
        exp_arr = exp_arr * ops.cast(directivity, "complex64")

        # Render pressure field for all relevant frequencies and sum them up
        rp = _pfield_freq_loop(
            freq,
            sound_speed,
            delays_tx,
            tx_apodization,
            exp_arr,
            exp_df,
            pulse_spect,
            probe_spect,
            z,
        )

        # RMS acoustic pressure
        p = ops.reshape(ops.sqrt(rp), size_downsampled)

        # resize p to exactly the original grid size
        p = ops.squeeze(ops.image.resize(p[..., None], size_orig, interpolation="nearest"), axis=-1)

        p_list.append(p)

        if verbose:
            progbar.add(1)

    p_arr = ops.convert_to_tensor(p_list)
    p_arr = ops.where(
        ops.isnan(p_arr), 0, p_arr
    )  # TODO: this is necessary for Jax / TF somehow. not sure why (not for torch)

    if norm:
        p_norm = normalize_pressure_field(p_arr, alpha=alpha, percentile=percentile)
    else:
        p_norm = p_arr

    return p_norm


def normalize_pressure_field(pfield, alpha: float = 1.0, percentile: float = 10.0):
    """
    Normalize the input array of intensities by zeroing out values below a given percentile.

    Args:
        pfield (array): The unnormalized pressure field array of shape (n_tx, Nz, Nx).
        alpha (float, optional): Exponent to 'sharpen or smooth' the weighting.
            Higher values result in sharper weighting. Default is 1.0.
        percentile (int, optional): minimum percentile threshold to keep in the weighting.
            Higher is more aggressive. Default is 10.

    Returns:
        ops.array: Normalized intensity array.
    """
    # Convert percentile to quantile (0–1 range)
    q = percentile / 100.0

    # Compute per-transmitter quantile thresholds
    threshold = ops.quantile(pfield, q, axis=(1, 2), keepdims=True)

    # Zero out values below the threshold
    pfield = ops.where(pfield < threshold, 0, pfield)

    # Sharpen the beam
    pfield = ops.power(pfield, alpha)

    # Normalize over transmit events (axis=0)
    p_norm = pfield / (keras.config.epsilon() + ops.sum(pfield, axis=0, keepdims=True))

    return p_norm


def _pfield_freq_step(
    k,
    freq,
    sound_speed,
    delays_tx,
    tx_apodization,
    rp_mono,
    pulse_spect,
    probe_spect,
    z,
):
    """
    Calculates the pressure field for a single frequency step.

    Args:
        k (int): Frequency index.
        freq (list): List of frequencies.
        sound_speed (float): Speed of sound.
        delays_tx (list): List of transmit delays.
        tx_apodization (list): List of transmit apodization values (complex64).
        rp_mono: (Tensor): Per-element, per-field-point complex pressure response
            (including directivity and propagation effects) at the current frequency sample.
        pulse_spect (list): List of pulse spectra.
        probe_spect (list): List of probe spectra (complex64).
        z (list): List of z-coordinates.

    Returns:
        rp_k (Tensor): Pressure field for this frequency.
    """
    kw = 2 * np.pi * freq[k] / sound_speed
    del_apod = ops.exp(1j * ops.cast(kw * sound_speed * delays_tx, "complex64")) * tx_apodization
    rp_k = ops.matmul(rp_mono, del_apod) * pulse_spect[k] * probe_spect[k]
    rp_k = ops.where(z < 0, 0, rp_k)
    return ops.abs(rp_k) ** 2


def _pfield_freq_loop(
    freq,
    sound_speed,
    delays_tx,
    tx_apodization,
    exp_arr,
    exp_df,
    pulse_spect,
    probe_spect,
    z,
):
    """Calculates the pressure field using frequency loop method.

    Args:
        freq (list): List of frequencies.
        sound_speed (float): Speed of sound.
        delays_tx (list): List of transmit delays.
        tx_apodization (list): List of transmit apodization values.
        exp_arr (list): List of complex exponentials.
        exp_df (list): List of complex exponential frequency shifts.
        pulse_spect (list): List of pulse spectra.
        probe_spect (list): List of probe spectra.
        z (list): List of z-coordinates.

    Returns:
        (Tensor): Pressure field.
    """

    tx_apodization = ops.cast(tx_apodization, "complex64")
    probe_spect = ops.cast(probe_spect, "complex64")
    rp_mono = exp_arr
    rp = 0
    for k in range(len(freq)):
        if k > 0:
            rp_mono *= exp_df
        rp_k = _pfield_freq_step(
            k,
            freq,
            sound_speed,
            delays_tx,
            tx_apodization,
            ops.mean(rp_mono, axis=1),  # avg over sub-elements
            pulse_spect,
            probe_spect,
            z,
        )
        rp += rp_k
    return rp



"""
Functions to write and validate datasets in the zea format.
"""

import inspect
from dataclasses import dataclass
from pathlib import Path

import numpy as np

from zea import log
from zea.data.file import File, validate_file
from zea.internal.checks import _DATA_TYPES
from zea.utils import first_not_none_item


@dataclass
class DatasetElement:
    """Class to store a dataset element with a name, data, description and unit. Used to
    supply additional dataset elements to the generate_zea_dataset function."""

    # The group name to store the dataset under. This can be a nested group, e.g.
    # "scan/waveforms"
    group_name: str
    # The name of the dataset. This will be the key in the group.
    dataset_name: str
    # The data to store in the dataset.
    data: np.ndarray
    description: str
    unit: str


def generate_example_dataset(
    path,
    add_optional_fields=False,
    add_optional_dtypes=False,
    n_frames=2,
    n_ax=2048,
    n_el=128,
    n_tx=11,
    n_ch=1,
    sound_speed=1540,
    center_frequency=7e6,
    sampling_frequency=40e6,
    n_z=512,
    n_x=512,
):
    """Generates an example dataset that contains all the necessary fields.
    Note: This dataset does not contain actual data, but is filled with random
    values.

    Args:
        path (str): The path to write the dataset to.
        add_optional_fields (bool, optional): Whether to add optional fields to
            the dataset. Defaults to False.
        add_optional_dtypes (bool, optional): Whether to add optional dtypes to
            the dataset. Defaults to False.
    """

    # creating some fake raw and image data
    raw_data = np.ones((n_frames, n_tx, n_ax, n_el, n_ch))
    # image data is in dB
    image = np.ones((n_frames, n_z, n_x)) * -40

    # creating some fake scan parameters
    t0_delays = np.zeros((n_tx, n_el), dtype=np.float32)
    tx_apodizations = np.zeros((n_tx, n_el), dtype=np.float32)
    probe_geometry = np.zeros((n_el, 3), dtype=np.float32)
    probe_geometry[:, 0] = np.linspace(-0.02, 0.02, n_el)
    initial_times = np.zeros((n_tx,), dtype=np.float32)

    if add_optional_fields:
        focus_distances = np.ones((n_tx,), dtype=np.float32) * np.inf
        tx_apodizations = np.zeros((n_tx, n_el), dtype=np.float32)
        polar_angles = np.zeros((n_tx,), dtype=np.float32)
        azimuth_angles = np.zeros((n_tx,), dtype=np.float32)
    else:
        focus_distances = None
        tx_apodizations = None
        polar_angles = None
        azimuth_angles = None

    if add_optional_dtypes:
        aligned_data = np.ones((n_frames, n_tx, n_ax, n_el, n_ch))
        envelope_data = np.ones((n_frames, n_z, n_x))
        beamformed_data = np.ones((n_frames, n_z, n_x, n_ch))
    else:
        aligned_data = None
        envelope_data = None
        beamformed_data = None

    generate_zea_dataset(
        path,
        raw_data=raw_data,
        aligned_data=aligned_data,
        envelope_data=envelope_data,
        beamformed_data=beamformed_data,
        image=image,
        probe_geometry=probe_geometry,
        sampling_frequency=sampling_frequency,
        center_frequency=center_frequency,
        initial_times=initial_times,
        t0_delays=t0_delays,
        sound_speed=sound_speed,
        tx_apodizations=tx_apodizations,
        probe_name="generic",
        focus_distances=focus_distances,
        polar_angles=polar_angles,
        azimuth_angles=azimuth_angles,
    )


def validate_input_data(raw_data, aligned_data, envelope_data, beamformed_data, image, image_sc):
    """
    Validates input data for generate_zea_dataset

    Args:
        raw_data (np.ndarray): The raw data of the ultrasound measurement of
            shape (n_frames, n_tx, n_ax, n_el, n_ch).
        aligned_data (np.ndarray): The aligned data of the ultrasound measurement of
            shape (n_frames, n_tx, n_ax, n_el, n_ch).
        envelope_data (np.ndarray): The envelope data of the ultrasound measurement of
            shape (n_frames, n_z, n_x).
        beamformed_data (np.ndarray): The beamformed data of the ultrasound measurement of
            shape (n_frames, n_z, n_x).
        image (np.ndarray): The ultrasound images to be saved of shape (n_frames, n_z, n_x).
        image_sc (np.ndarray): The scan converted ultrasound images to be saved
            of shape (n_frames, output_size_z, output_size_x).
    """
    assert (
        raw_data is not None
        or aligned_data is not None
        or envelope_data is not None
        or beamformed_data is not None
        or image is not None
        or image_sc is not None
    ), f"At least one of the data types {_DATA_TYPES} must be specified."

    # specific checks for each data type are done in validate_file


def _write_datasets(
    dataset,
    data_group_name="data",
    scan_group_name="scan",
    raw_data=None,
    aligned_data=None,
    envelope_data=None,
    beamformed_data=None,
    image=None,
    image_sc=None,
    n_ax=None,
    n_el=None,
    n_tx=None,
    n_ch=None,
    n_frames=None,
    sound_speed=None,
    probe_geometry=None,
    sampling_frequency=None,
    center_frequency=None,
    initial_times=None,
    t0_delays=None,
    tx_apodizations=None,
    focus_distances=None,
    polar_angles=None,
    azimuth_angles=None,
    bandwidth_percent=None,
    time_to_next_transmit=None,
    tgc_gain_curve=None,
    element_width=None,
    tx_waveform_indices=None,
    waveforms_one_way=None,
    waveforms_two_way=None,
    additional_elements=None,
    cast_to_float=True,
    **kwargs,
):
    if kwargs:
        raise ValueError(f"Unknown arguments: {list(kwargs.keys())}")

    def _convert_datatype(x, astype=np.float32):
        if cast_to_float:
            return x.astype(astype) if x is not None else None
        else:
            return x

    def _first_not_none_shape(arr, axis):
        data = first_not_none_item(arr)
        return data.shape[axis] if data is not None else None

    def _add_dataset(group_name: str, name: str, data: np.ndarray, description: str, unit: str):
        """Adds a dataset to the given group with a description and unit.
        If data is None, the dataset is not added."""
        if data is None:
            return

        # Create the group if it does not exist
        if group_name not in dataset:
            group = dataset.create_group(group_name)
        else:
            group = dataset[group_name]

        new_dataset = group.create_dataset(name, data=data)
        new_dataset.attrs["description"] = description
        new_dataset.attrs["unit"] = unit

    # Write data group
    data_group = dataset.create_group(data_group_name)
    data_group.attrs["description"] = "This group contains the data."

    if n_frames is None:
        n_frames = first_not_none_item(
            [raw_data, aligned_data, envelope_data, beamformed_data, image, image_sc]
        ).shape[0]
    if n_tx is None:
        n_tx = _first_not_none_shape([raw_data, aligned_data], axis=1)
    if n_ax is None:
        n_ax = _first_not_none_shape([raw_data, aligned_data, beamformed_data], axis=-3)
    if n_ax is None:
        n_ax = _first_not_none_shape([envelope_data, image, image_sc], axis=-2)
    if n_el is None:
        n_el = _first_not_none_shape([raw_data], axis=-2)
    if n_ch is None:
        n_ch = _first_not_none_shape([raw_data, aligned_data, beamformed_data], axis=-1)
    if n_tx is None:
        n_tx = _first_not_none_shape([t0_delays, focus_distances, polar_angles], axis=0)
    if n_el is None:
        n_el = _first_not_none_shape([t0_delays], axis=1)
    if n_el is None:
        n_el = _first_not_none_shape([probe_geometry], axis=0)
    if n_tx is None:
        n_tx = 1

    _add_dataset(
        group_name=data_group_name,
        name="raw_data",
        data=_convert_datatype(raw_data),
        description="The raw_data of shape (n_frames, n_tx, n_ax, n_el, n_ch).",
        unit="unitless",
    )

    _add_dataset(
        group_name=data_group_name,
        name="aligned_data",
        data=_convert_datatype(aligned_data),
        description="The aligned_data of shape (n_frames, n_tx, n_ax, n_el, n_ch).",
        unit="unitless",
    )

    _add_dataset(
        group_name=data_group_name,
        name="envelope_data",
        data=_convert_datatype(envelope_data),
        description="The envelope_data of shape (n_frames, n_z, n_x).",
        unit="unitless",
    )

    _add_dataset(
        group_name=data_group_name,
        name="beamformed_data",
        data=_convert_datatype(beamformed_data),
        description="The beamformed_data of shape (n_frames, n_z, n_x).",
        unit="unitless",
    )

    _add_dataset(
        group_name=data_group_name,
        name="image",
        data=_convert_datatype(image),
        unit="unitless",
        description="The images of shape [n_frames, n_z, n_x]",
    )

    _add_dataset(
        group_name=data_group_name,
        name="image_sc",
        data=_convert_datatype(image_sc),
        unit="unitless",
        description=("The scan converted images of shape [n_frames, output_size_z, output_size_x]"),
    )

    # Write scan group
    scan_group = dataset.create_group(scan_group_name)
    scan_group.attrs["description"] = "This group contains the scan parameters."

    _add_dataset(
        group_name=scan_group_name,
        name="n_ax",
        data=n_ax,
        description="The number of axial samples.",
        unit="unitless",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="n_el",
        data=n_el,
        description="The number of elements in the probe.",
        unit="unitless",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="n_tx",
        data=n_tx,
        description="The number of transmits per frame.",
        unit="unitless",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="n_ch",
        data=n_ch,
        description=("The number of channels. For RF data this is 1. For IQ data this is 2."),
        unit="unitless",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="n_frames",
        data=n_frames,
        description="The number of frames.",
        unit="unitless",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="sound_speed",
        data=sound_speed,
        description="The speed of sound in m/s",
        unit="m/s",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="probe_geometry",
        data=probe_geometry,
        description="The probe geometry of shape (n_el, 3).",
        unit="m",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="sampling_frequency",
        data=sampling_frequency,
        description="The sampling frequency in Hz.",
        unit="Hz",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="center_frequency",
        data=center_frequency,
        description="The center frequency in Hz.",
        unit="Hz",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="initial_times",
        data=initial_times,
        description=(
            "The times when the A/D converter starts sampling "
            "in seconds of shape (n_tx,). This is the time between the "
            "first element firing and the first recorded sample."
        ),
        unit="s",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="t0_delays",
        data=t0_delays,
        description="The t0_delays of shape (n_tx, n_el).",
        unit="s",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="tx_apodizations",
        data=tx_apodizations,
        description=(
            "The transmit delays for each element defining the"
            " wavefront in seconds of shape (n_tx, n_elem). This is"
            " the time at which each element fires shifted such that"
            " the first element fires at t=0."
        ),
        unit="unitless",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="focus_distances",
        data=focus_distances,
        description=(
            "The transmit focus distances in meters of "
            "shape (n_tx,). For planewaves this is set to Inf."
        ),
        unit="m",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="polar_angles",
        data=polar_angles,
        description=("The polar angles of the transmit beams in radians of shape (n_tx,)."),
        unit="rad",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="azimuth_angles",
        data=azimuth_angles,
        description=("The azimuthal angles of the transmit beams in radians of shape (n_tx,)."),
        unit="rad",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="bandwidth_percent",
        data=bandwidth_percent,
        description=("The receive bandwidth of RF signal in percentage of center frequency."),
        unit="unitless",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="time_to_next_transmit",
        data=time_to_next_transmit,
        description=("The time between subsequent transmit events of shape (n_frames, n_tx)."),
        unit="s",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="tgc_gain_curve",
        data=tgc_gain_curve,
        description=(
            "The time-gain-compensation that was applied to every sample in the "
            "raw_data of shape (n_ax,). Divide by this curve to undo the TGC."
        ),
        unit="unitless",
    )

    _add_dataset(
        group_name=scan_group_name,
        name="element_width",
        data=element_width,
        description="The width of the elements in the probe in meters.",
        unit="m",
    )

    if tx_waveform_indices is not None and (
        waveforms_one_way is not None or waveforms_two_way is not None
    ):
        _add_dataset(
            group_name=scan_group_name,
            name="tx_waveform_indices",
            data=tx_waveform_indices,
            description=(
                "Transmit indices for waveforms, indexing waveforms_one_way "
                "and waveforms_two_way. This indicates which transmit waveform was "
                "used for each transmit event."
            ),
            unit="-",
        )
        n_waveforms = len(waveforms_one_way)
        for n, waveform_1way, waveform_2way in zip(
            range(n_waveforms), waveforms_one_way, waveforms_two_way
        ):
            _add_dataset(
                group_name=scan_group_name + "/waveforms_one_way",
                name=f"waveform_{str(n).zfill(3)}",
                data=waveform_1way,
                description=(
                    "One-way waveform as simulated by the Verasonics system, "
                    "sampled at 250MHz. This is the waveform after being filtered "
                    "by the tranducer bandwidth once."
                ),
                unit="V",
            )
            _add_dataset(
                group_name=scan_group_name + "/waveforms_two_way",
                name=f"waveform_{str(n).zfill(3)}",
                data=waveform_2way,
                description=(
                    "Two-way waveform as simulated by the Verasonics system, "
                    "sampled at 250MHz. This is the waveform after being filtered "
                    "by the tranducer bandwidth twice."
                ),
                unit="V",
            )

    # Add additional elements
    if additional_elements is not None:
        for element in additional_elements:
            _add_dataset(
                group_name=element.group_name,
                name=element.dataset_name,
                data=element.data,
                description=element.description,
                unit=element.unit,
            )


def generate_zea_dataset(
    path,
    raw_data=None,
    aligned_data=None,
    envelope_data=None,
    beamformed_data=None,
    image=None,
    image_sc=None,
    probe_geometry=None,
    sampling_frequency=None,
    center_frequency=None,
    initial_times=None,
    t0_delays=None,
    sound_speed=None,
    probe_name=None,
    description="No description was supplied",
    focus_distances=None,
    polar_angles=None,
    azimuth_angles=None,
    tx_apodizations=None,
    bandwidth_percent=None,
    time_to_next_transmit=None,
    tgc_gain_curve=None,
    element_width=None,
    tx_waveform_indices=None,
    waveforms_one_way=None,
    waveforms_two_way=None,
    additional_elements=None,
    event_structure=False,
    cast_to_float=True,
):
    """Generates a dataset in the zea format.

    Args:
        path (str): The path to write the dataset to.
        raw_data (np.ndarray): The raw data of the ultrasound measurement of
            shape (n_frames, n_tx, n_ax, n_el, n_ch).
        aligned_data (np.ndarray): The aligned data of the ultrasound measurement of
            shape (n_frames, n_tx, n_ax, n_el, n_ch).
        envelope_data (np.ndarray): The envelope data of the ultrasound measurement of
            shape (n_frames, n_z, n_x).
        beamformed_data (np.ndarray): The beamformed data of the ultrasound measurement of
            shape (n_frames, n_z, n_x, n_ch).
        image (np.ndarray): The ultrasound images to be saved of shape (n_frames, n_z, n_x).
        image_sc (np.ndarray): The scan converted ultrasound images to be saved
            of shape (n_frames, output_size_z, output_size_x).
        probe_geometry (np.ndarray): The probe geometry of shape (n_el, 3).
        sampling_frequency (float): The sampling frequency in Hz.
        center_frequency (float): The center frequency in Hz.
        initial_times (list): The times when the A/D converter starts sampling
            in seconds of shape (n_tx,). This is the time between the first element
            firing and the first recorded sample.
        t0_delays (np.ndarray): The t0_delays of shape (n_tx, n_el).
        sound_speed (float): The speed of sound in m/s.
        probe_name (str): The name of the probe.
        description (str): The description of the dataset.
        focus_distances (np.ndarray): The focus distances of shape (n_tx, n_el).
        polar_angles (np.ndarray): The polar angles (radians) of shape (n_el,).
        azimuth_angles (np.ndarray): The azimuth angles (radians) of shape (n_tx,).
        tx_apodizations (np.ndarray): The transmit delays for each element defining
            the wavefront in seconds of shape (n_tx, n_elem).
            This is the time between the first element firing and the last element firing.
        bandwidth_percent (float): The bandwidth of the transducer as a
            percentage of the center frequency.
        time_to_next_transmit (np.ndarray): The time between subsequent transmit events in s
            of shape (n_frames, n_tx).
        tgc_gain_curve (np.ndarray): The TGC gain that was applied to every sample in the
            raw_data of shape (n_ax).
        element_width (float): The width of the elements in the probe in meters of
            shape (n_tx,).
        tx_waveform_indices (np.ndarray): Transmit indices for waveforms, indexing
            waveforms_one_way and waveforms_two_way. This indicates which transmit
            waveform was used for each transmit event.
        waveforms_one_way (list): List of one-way waveforms as simulated by the Verasonics
            system, sampled at 250MHz. This is the waveform after being filtered by the
            tranducer bandwidth once. Every element in the list is a 1D numpy array.
        waveforms_two_way (list): List of two-way waveforms as simulated by the Verasonics
            system, sampled at 250MHz. This is the waveform after being filtered by the
            tranducer bandwidth twice. Every element in the list is a 1D numpy array.
        additional_elements (List[DatasetElement]): A list of additional dataset
            elements to be added to the dataset. Each element should be a DatasetElement
            object. The additional elements are added under the scan group.
        event_structure (bool): Whether to write the dataset with an event structure.
            In that case all data should be lists with the same length (number of events).
            The data will be stored under event_i/data and event_i/scan for each event i.
            Instead of just a single data and scan group.
        cast_to_float (bool): Whether to store data as float32. You may want to set this
            to False if storing images.

    """
    # check if all args are lists
    if isinstance(probe_name, list):
        # all names in probe_name list should be the same
        assert len(set(probe_name)) == 1, "Probe names for all events should be the same"

    data_and_parameters = {
        "raw_data": raw_data,
        "aligned_data": aligned_data,
        "envelope_data": envelope_data,
        "beamformed_data": beamformed_data,
        "image": image,
        "image_sc": image_sc,
        "probe_geometry": probe_geometry,
        "sampling_frequency": sampling_frequency,
        "center_frequency": center_frequency,
        "initial_times": initial_times,
        "t0_delays": t0_delays,
        "sound_speed": sound_speed,
        "probe_name": probe_name,
        "description": description,
        "focus_distances": focus_distances,
        "polar_angles": polar_angles,
        "azimuth_angles": azimuth_angles,
        "tx_apodizations": tx_apodizations,
        "bandwidth_percent": bandwidth_percent,
        "time_to_next_transmit": time_to_next_transmit,
        "tgc_gain_curve": tgc_gain_curve,
        "element_width": element_width,
        "tx_waveform_indices": tx_waveform_indices,
        "waveforms_one_way": waveforms_one_way,
        "waveforms_two_way": waveforms_two_way,
        "additional_elements": additional_elements,
    }

    # make sure input arguments of func is same length as data_and_parameters
    # except `path` and `event_structure` arguments and ofcourse `data_and_parameters` itself
    assert (
        len(data_and_parameters) == len(inspect.signature(generate_zea_dataset).parameters) - 3
    ), (
        "All arguments should be put in data_and_parameters except "
        "`path`, `event_structure`, and `cast_to_float` arguments."
    )

    if event_structure:
        for argument, argument_value in data_and_parameters.items():
            _num_events = None
            if argument_value is not None:
                assert isinstance(argument_value, list), (
                    f"{argument} should be a list when event_structure is set to True."
                )
                num_events = len(argument_value)
                if _num_events is not None:
                    assert num_events == _num_events, (
                        "All arguments should have the same number of events."
                    )
                _num_events = num_events

        assert len(set(probe_name)) == 1, "Probe names for all events should be the same"
        log.info(
            f"Event structure is set to True. Writing dataset with event "
            f"structure (found {len(probe_name)} events)."
        )
        num_events = len(probe_name)
        probe_name = probe_name[0]
        description = description[0]

    assert isinstance(probe_name, str), "The probe name must be a string."
    assert isinstance(description, str), "The description must be a string."
    assert isinstance(event_structure, bool), "The event_structure must be a boolean."

    validate_input_data(
        raw_data=raw_data,
        aligned_data=aligned_data,
        envelope_data=envelope_data,
        beamformed_data=beamformed_data,
        image=image,
        image_sc=image_sc,
    )

    # Convert path to Path object
    path = Path(path)

    if path.exists():
        raise FileExistsError(f"The file {path} already exists.")

    # Create the directory if it does not exist
    path.parent.mkdir(parents=True, exist_ok=True)

    with File(path, "w") as dataset:
        dataset.attrs["probe"] = probe_name
        dataset.attrs["description"] = description
        dataset.attrs["event_structure"] = event_structure
        # remove probe and description from data_and_parameters
        data_and_parameters.pop("probe_name")
        data_and_parameters.pop("description")

        if event_structure:
            for i in range(num_events):
                _data_and_parameters = {
                    k: v[i] for k, v in data_and_parameters.items() if v is not None
                }
                _write_datasets(
                    dataset,
                    data_group_name=f"event_{i}/data",
                    scan_group_name=f"event_{i}/scan",
                    cast_to_float=cast_to_float,
                    **_data_and_parameters,
                )

        else:
            _write_datasets(
                dataset,
                data_group_name="data",
                scan_group_name="scan",
                cast_to_float=cast_to_float,
                **data_and_parameters,
            )

    validate_file(path)
    log.info(f"zea dataset written to {log.yellow(path)}")



"""Lens corrected delay computation for ultrasound beamforming."""

from keras import ops


def calculate_lens_corrected_delays(
    grid,
    t0_delays,
    tx_apodizations,
    probe_geometry,
    initial_times,
    sampling_frequency,
    sound_speed,
    n_tx,
    n_el,
    focus_distances,
    polar_angles,
    lens_sound_speed=1000,
    lens_thickness=1e-3,
    n_iter=2,
):
    """Compute the lens corrected delays IN SAMPLES for a given grid and transmit scheme.

    Args:
        grid (ndarray): The grid of shape (n_pixels, 3).
        t0_delays (ndarray): The t0 delays of shape (n_tx, n_el).
        tx_apodizations (ndarray): The transmit apodizations of shape (n_tx, n_el).
        probe_geometry (ndarray): The probe geometry of shape (n_el, 3).
        initial_times (ndarray): The initial times of shape (n_tx,).
        sampling_frequency (float): The sampling frequency in Hz.
        sound_speed (float): The speed of sound in the medium in m/s.
        n_tx (int): The number of transmits.
        n_el (int): The number of elements.
        focus_distances (ndarray): The focus distances of shape (n_tx,).
        polar_angles (ndarray): The polar angles of shape (n_tx,).
        lens_sound_speed (float): The speed of sound in the lens in m/s.
        lens_thickness (float): The thickness of the lens in meters.
        n_iter (int): The number of iterations to run the Newton-Raphson method.

    Returns:
        ndarray: The transmit delays in samples of shape (n_pixels, n_tx).
        ndarray: The receive delays in samples of shape (n_pixels, n_el).
    """
    for arr in [t0_delays, grid, tx_apodizations, probe_geometry]:
        assert arr.ndim == 2
    assert probe_geometry.shape[0] == n_el
    assert t0_delays.shape[0] == n_tx

    tx_delays = []
    rx_delays = compute_lens_corrected_travel_times(
        probe_geometry,
        grid,
        lens_thickness,
        lens_sound_speed,
        sound_speed,
        n_iter=n_iter,
    )
    for tx in range(n_tx):
        # Add a large offset to elements that are not used in the transmit to
        # diqualify them from being the closest element
        apod_offset = ops.where(tx_apodizations[tx] == 0, 10.0, 0)
        tx_min = ops.min(rx_delays + t0_delays[tx] + apod_offset, axis=-1) + initial_times[tx]
        tx_delays.append(tx_min)
    tx_delays = ops.stack(tx_delays, axis=-1)

    tx_delays *= sampling_frequency
    rx_delays *= sampling_frequency

    return tx_delays, rx_delays


def compute_lens_corrected_travel_times(
    element_pos, pixel_pos, lens_thickness, c_lens, c_medium, n_iter=1
):
    """Compute the travel time of the shortest path between the element and the pixel.

    Args:
        element_pos (ndarray): The position of the element of shape (n_elements, 3).
        pixel_pos (ndarray): The position of the pixel of shape (n_pixels, 3).
        lens_thickness (float): The thickness of the lens in meters.
        c_lens (float): The speed of sound in the lens in m/s.
        c_medium (float): The speed of sound in the medium in m/s.
        n_iter (int): The number of iterations to run the Newton-Raphson method.

    Returns:
        ndarray: The travel times of shape (n_pixels, n_elements).
    """

    pixel_pos = pixel_pos[:, None] - element_pos[None]

    # Project the 3D problem to a 2D problem by shifting all pixels to have the element
    # in the origin and then projecting the positions to the plane spanned by
    # [pixel_x-element_x, pixel_z-element_z, 0], [0, 0, 1]
    xs = ops.norm(pixel_pos[..., :2], axis=-1)
    zs = pixel_pos[..., -1]

    pixel_pos_2d = ops.stack([xs, zs], axis=-1)
    element_pos_2d = ops.zeros((1, element_pos.shape[0], 2))

    xl = compute_xl(
        element_pos_2d,
        pixel_pos_2d,
        lens_thickness,
        c_lens,
        c_medium,
        n_iter,
    )

    # Form the position of the lens crossing point by adding the z-coordinate of the
    # lens to the point
    pos_lenscrossing = ops.stack([xl, lens_thickness * ops.ones_like(xl)], axis=-1)

    indices = ops.array([0, -1])
    element_pos = ops.take(element_pos, indices, axis=-1)

    # Compute the travel time of the shortest path
    travel_time = compute_travel_time(
        element_pos_2d, pos_lenscrossing, c_lens
    ) + compute_travel_time(pos_lenscrossing, pixel_pos_2d, c_medium)
    return travel_time


def compute_xl(element_pos_2d, pixel_pos_2d, lens_thickness, c_lens, c_medium, n_iter):
    """Computes the lateral point on the lens that the shortest path goes through based
    on Fermat's principle.

    Args:
        element_pos_2d (float): The 2D position of the element.
        pixel_pos_2d (float): The 2D position of the pixel.
        lens_thickness (float): The thickness of the lens in meters.
        c_lens (float): The speed of sound in the lens in m/s.
        c_medium (float): The speed of sound in the medium in m/s.
        n_iter (int): The number of iterations to run the Newton-Raphson method.

    Returns:
        float: The x-coordinate of the lateral point on the lens.
    """
    xs = pixel_pos_2d[..., 0]
    zs = pixel_pos_2d[..., 1]
    xe = element_pos_2d[..., 0]
    ze = element_pos_2d[..., 1]

    # Apply Newton-Raphson method to find the lateral point on the lens that the shortest
    # path goes through
    xl_init = lens_thickness * (xs - xe) / (zs - ze) + xe
    xl = xl_init
    for _ in range(n_iter):
        xl = xl + dxl(xe, ze, xl, xs, zs, lens_thickness, c_lens, c_medium)

        # Clip the lateral point to be in between the element and the pixel
        xl = ops.clip(xl, ops.minimum(xs, xe), ops.maximum(xs, xe))

    return xl


def dxl(xe, ze, xl, xs, zs, zl, c_lens, c_medium):
    """Computes the update step for the lateral point on the lens that the shortest path
    using the Newton-Raphson method.

    Notes
    -----
    This result was derived by defining the total travel time through the lens and the
    medium as a function of the lateral point on the lens and then taking the
    derivative. We then have a function whose root is the lateral point on the lens that
    the shortest path goes through. We then compute the derivative and update the
    lateral point on the lens using the Newton-Raphson method:
    x_new = x - f(x) / f'(x).
    """

    eps = 1e-6

    numerator = -((xe - xl) / (c_lens * ops.sqrt((xe - xl) ** 2 + (ze - zl) ** 2))) + (
        (xl - xs) / (c_medium * ops.sqrt((xl - xs) ** 2 + (zl - zs) ** 2)) + eps
    )

    denominator = (
        -((xe - xl) ** 2 / (c_lens * ((xe - xl) ** 2 + (ze - zl) ** 2) ** (3 / 2) + eps))
        + (1 / (c_lens * ops.sqrt((xe - xl) ** 2 + (ze - zl) ** 2)))
        - ((xl - xs) ** 2 / (c_medium * ((xl - xs) ** 2 + (zl - zs) ** 2) ** (3 / 2) + eps))
        + (1 / (c_medium * ops.sqrt((xl - xs) ** 2 + (zl - zs) ** 2) + eps))
    )

    result = -numerator / (denominator + eps)

    # Handle NaNs
    result = ops.nan_to_num(result)

    # Clip the update step to prevent divergence
    # This value is chosen to be small enough to prevent divergence but large enough to
    # cover the distance accross a normal ultrasound aperture in a single step.
    result = ops.clip(result, -10e-3, 10e-3)

    return result


def compute_travel_time(pos_a, pos_b, c):
    """Compute the travel time between two points."""
    return ops.linalg.norm(pos_a - pos_b, axis=-1) / c



"""Config module for managing configuration settings.

This module provides the :class:`Config` class for managing configuration settings,
with support for loading from YAML files, HuggingFace Hub, and dot notation access.

Features
--------

- Dot notation access to dictionary keys.
- Recursive conversion of nested dictionaries/lists to Config objects.
- Attribute access logging and suggestion of similar attribute names.
- Freezing/unfreezing to prevent/allow new attributes.
- Serialization to YAML/JSON.
- Integration with Hugging Face Hub.

Example Usage
^^^^^^^^^^^^^

.. code-block:: python

    from zea import Config

    # Load from YAML
    config = Config.from_yaml("config.yaml")
    # Load from HuggingFace Hub
    config = Config.from_hf("zea/diffusion-echonet-dynamic", "train_config.yaml")

    # Access attributes with dot notation
    print(config.model.name)

    # Update recursively
    config.update_recursive({"model": {"name": "new_model"}})

    # Save to YAML
    config.save_to_yaml("new_config.yaml")

"""

import copy
import difflib
import inspect
import json
from pathlib import Path
from typing import Union

import yaml
from huggingface_hub import hf_hub_download

from zea import log
from zea.internal.config.validation import config_schema
from zea.internal.core import object_to_tensor


class Config(dict):
    """Config class.

    This Config class extends a normal dictionary with dot notation access.

    Features:
        - `Config.from_yaml` method to load a config from a yaml file.
        - `Config.from_hf` method to load a config from a huggingface hub.
        - `save_to_yaml` method to save the config to a yaml file.
        - `copy` method to create a deep copy of the config.
        - Normal dictionary methods such as `keys`, `values`, `items`, `pop`, `update`, `get`.
        - Propose similar attribute names if a non-existing attribute is accessed.
        - Freeze the config object to prevent new attributes from being added.
        - Load config object from yaml file.
        - Logs all accessed attributes such that you can check if all attributes have been accessed.

    We took inspiration from the following sources:
        - `EasyDict <https://pypi.org/project/easydict/>`_
        - `keras.utils.Config <https://keras.io/api/utils/experiment_management_utils/#config-class>`_

    But this implementation is superior :)

    """  # noqa: E501

    __frozen__ = False

    def __init__(self, dictionary=None, __parent__=None, **kwargs):
        """
        Initializes a Config object.

        Args:
            dictionary (dict, optional): A dictionary containing key-value pairs
                to initialize the Config object. Defaults to None.
            **kwargs: Additional key-value pairs to initialize the Config object.
                Will override values in the dictionary if they have the same key.
        """
        # Get all methods of the Config class and store them in a list as protected attributes
        super().__setattr__(
            "__protected__",
            [x[0] for x in inspect.getmembers(Config, predicate=inspect.isroutine)]
            + ["__protected__", "__accessed__", "__parent__"],
        )
        super().__setattr__("__accessed__", {})
        super().__setattr__("__parent__", __parent__)

        if dictionary is None:
            dictionary = {}
        if kwargs:
            dictionary.update(**kwargs)
        for k, v in dictionary.items():
            setattr(self, k, v)

    def clear(self):
        """Clear the config object."""
        super().clear()
        self._reset_accessed()

    def fromkeys(self, keys, value=None):
        """Returns a config with the specified keys and value"""
        super().fromkeys(keys, value)

    def get(self, key, default=None):
        """Returns the value of the specified key"""
        self._mark_accessed(key)
        return super().get(key, default)

    def items(self):
        """Returns a list containing a tuple for each key value pair"""
        # Use a generator that calls __getitem__ for every key
        return [(key, self[key]) for key in self.keys()]

    def keys(self):
        """Returns a list containing the config's keys"""
        return super().keys()

    def pop(self, key, default=None):
        """Removes the element with the specified key"""
        self._mark_accessed(key)
        return super().pop(key, default)

    def popitem(self):
        """Removes the last inserted key-value pair"""
        key, value = super().popitem()
        self._mark_accessed(key)
        return key, value

    def setdefault(self, key, default=None):
        """Returns the value of the specified key. If the key does not exist: insert the key,
        with the specified value"""
        # Use __getitem__ to get values and __setitem__ to set values
        if key not in self:
            self[key] = default
        return self[key]

    def update(self, dictionary: dict | None = None, **kwargs):
        """Updates the config with the specified key-value pairs"""
        # Use __setitem__ to set values
        if dictionary is None:
            dictionary = {}
        dictionary.update(kwargs)
        for key, value in dictionary.items():
            self[key] = value

    def update_recursive(self, dictionary: dict | None = None, **kwargs):
        """Recursively update the config with the provided dictionary and keyword arguments.

        If a key corresponds to another Config object, the update_recursive
        method is called recursively on that object. This makes it possible
        to update nested Config objects without replacing them.

        If a value is a list and the corresponding config value is also a list,
        each element is updated recursively if it is a Config, otherwise replaced.

        Example:

        .. code-block:: python

            config = Config({"a": 1, "b": {"c": 2, "d": 3}})
            config.update_recursive({"a": 4, "b": {"c": 5}})
            print(config)
            # <Config {'a': 4, 'b': {'c': 5, 'd': 3}}>
            # Notice how "d" is kept and only "c" is updated.

        Args:
            dictionary (dict, optional): Dictionary to update from.
            **kwargs: Additional key-value pairs to update.
        """
        if dictionary is None:
            dictionary = {}
        dictionary.update(kwargs)
        for key, value in dictionary.items():
            if key in self and isinstance(self[key], Config):
                self[key].update_recursive(value)
            elif key in self and isinstance(value, list):
                for i, v in enumerate(value):
                    if isinstance(v, Config):
                        self[key][i].update_recursive(v)
                    else:
                        self[key][i] = v
            else:
                self[key] = value

    def values(self):
        """Returns a list of all the values in the config"""
        # Use __getitem__ to get values
        return (self[key] for key in self.keys())

    def __or__(self, other):
        """
        `self | other` operator. Returns a new config object with the contents of both configs.
        """
        return Config(super().__or__(other))

    def __ror__(self, other):
        """
        `other | self` operator. Returns a new config object with the contents of both configs.
        """
        return Config(super().__ror__(other))

    def __ior__(self, other):
        """
        `|=` operator. Updates the config with the contents of the other config.
        """
        self.update(other)
        return self

    def __iter__(self):
        """Returns an iterator that iterates through the keys of the config"""
        # Overwritten to ensure iteration respects our logic
        return iter(self.keys())

    def __contains__(self, key):
        """Returns True if the specified key exists in the config"""
        return super().__contains__(key)

    def __setattr__(self, name, value):
        # Check if attribute is a method of the Config class, this cannot be overridden
        if hasattr(self, "__protected__") and name in self.__protected__:
            raise AttributeError(f"Cannot set attribute `{name}`. It is used by the Config class.")

        # Check if config is frozen
        if self.__frozen__ and not hasattr(self, name):
            raise TypeError(
                f"Config is a frozen, no new attributes can be added. Tried to add: `{name}`"
            )

        # If overriding an existing attribute, mark it as unaccessed
        self._mark_unaccessed(name)

        # Convert tuple to list to allow for item assignment
        if isinstance(value, tuple):
            value = list(value)

        # Ensures lists and tuples of dictionaries are converted to Config objects as well
        if isinstance(value, list):
            value = [
                self.__class__(x, __parent__=self) if isinstance(x, dict) else x for x in value
            ]
        # Ensures dictionaries are converted to Config objects as well
        elif isinstance(value, dict):
            value = self.__class__(value, __parent__=self)

        super().__setitem__(name, value)

    def __setitem__(self, key, value):
        return self.__setattr__(key, value)

    def _unknown_attr(self, name):
        msg = f"Unknown attribute: '{name}'."
        if "difflib" in globals():
            closest_matches = difflib.get_close_matches(name, self.keys(), n=1, cutoff=0.7)
            if closest_matches:
                msg += f" Did you mean '{closest_matches[0]}'?"
        return msg

    def _reset_accessed(self):
        """Reset accessed attributes."""
        self._recursive_setattr("__accessed__", {})

    def _mark_accessed(self, name):
        """Mark an attribute as accessed."""
        if name in self:
            self.__accessed__[name] = True

    def _mark_unaccessed(self, name):
        """Mark an attribute as unaccessed."""
        if name in self.__accessed__:
            del self.__accessed__[name]

    def _mark_accessed_recursive(self):
        """Mark an attribute and all its children as accessed."""

        def mark_accessed(config, key, value):
            config._mark_accessed(key)
            return key, value

        self.as_dict(mark_accessed)

    def _dict_items(self):
        """Return the items of the config object. Only used for internal purposes."""
        return super().items()

    def _trace_through_ancestors(self, key_trace=None):
        """Find the root ancestor of the config object."""
        if key_trace is None:
            key_trace = []
        if self.__parent__ is None:
            return self, key_trace
        for key, value in self.__parent__._dict_items():
            if isinstance(value, list):
                for i, v in enumerate(value):
                    if v == self:
                        return self.__parent__._trace_through_ancestors([key + f"_{i}"] + key_trace)
            if value == self:
                return self.__parent__._trace_through_ancestors([key] + key_trace)
        raise ValueError("Parent not found in ancestors. Report to zea developers.")

    @staticmethod
    def _assert_key_accessed(config, key, value, _assert=True):
        """Assert that a key has been accessed."""
        if key not in config.__accessed__:
            key_trace = config._trace_through_ancestors()[1]
            msg = f"Attribute '{key}'='{value}' has not been accessed."
            if key_trace:
                msg += f" Has ancestors through '{key_trace}'"
            if _assert:
                raise AssertionError(msg)
            log.warning(msg)
        return key, value

    def _assert_all_accessed(self):
        """Assert that all attributes have been accessed."""
        # Temporary remove parent to avoid recursion if not being called from ancestor.
        self._all_unaccessed(_assert=True)

    def _log_all_unaccessed(self):
        """Log all unaccessed attributes."""
        self._all_unaccessed(_assert=False)

    def _all_unaccessed(self, _assert=False):
        """Assert or log all unaccessed attributes."""
        # Temporary remove parent to avoid recursion if not being called from ancestor.
        parent = self.__parent__
        super().__setattr__("__parent__", None)
        self.as_dict(lambda *args: self._assert_key_accessed(*args, _assert=_assert))
        super().__setattr__("__parent__", parent)

    def __getattr__(self, name):
        if name in self:
            self._mark_accessed(name)
            return super().__getitem__(name)

        msg = self._unknown_attr(name)
        raise AttributeError(msg)

    def __getitem__(self, key):
        if key in self:
            self._mark_accessed(key)
            return super().__getitem__(key)

        msg = self._unknown_attr(key)
        raise KeyError(msg)

    def __delattr__(self, name):
        del self[name]

    def __repr__(self):
        return f"<Config {self.as_dict()}>"

    def to_json(self):
        """Return the config as a json string."""
        return json.dumps(self)

    def as_dict(self, func_on_leaves=None):
        """Convert the config to a normal dictionary (recursively).

        Args:
            func_on_leaves (callable, optional): Function to apply to each leaf node.
                The function should take three arguments: the config object, the key, and the value.
                You can change the key and value inside the function. Defaults to None.
        """
        dictionary = {}
        for key, value in self._dict_items():
            if isinstance(value, Config):
                value = value.as_dict(func_on_leaves)
            elif isinstance(value, list):
                value = [v.as_dict(func_on_leaves) if isinstance(v, Config) else v for v in value]
            # a dict does not exist inside a Config object, because it is a Config object itself
            if func_on_leaves:
                key, value = func_on_leaves(self, key, value)
            dictionary[key] = value
        return dictionary

    def serialize(self):
        """Return a dict of this config object with all Path objects converted to strings."""
        return self.as_dict(lambda _, key, value: (key, _path_to_str(value)))

    def copy(self):
        """Deep copy the config object.

        This is useful when you want to modify the config object
        without changing the original. Does not preserve the access history or frozen state!
        """
        return Config(copy.deepcopy(self.as_dict()))

    def save_to_yaml(self, path):
        """Save config contents to yaml"""
        with open(Path(path), "w", encoding="utf-8") as save_file:
            yaml.dump(
                self.serialize(),
                save_file,
                default_flow_style=False,
                sort_keys=False,
            )

    def freeze(self):
        """Freeze config object.

        This means that no new attributes can be added.
        Only existing attributes can be modified.
        """
        self._recursive_setattr("__frozen__", True)

    def unfreeze(self):
        """Unfreeze config object. This means that new attributes can be added."""
        self._recursive_setattr("__frozen__", False)

    def _recursive_setattr(self, set_key, set_value):
        """Helper function to recursively set an attribute on all nested configs."""
        super().__setattr__(set_key, set_value)
        for _, value in self._dict_items():
            if isinstance(value, Config):
                value._recursive_setattr(set_key, set_value)
            elif isinstance(value, list):
                for v in value:
                    if isinstance(v, Config):
                        v._recursive_setattr(set_key, set_value)

    @classmethod
    def from_yaml(cls, path, **kwargs):
        """Load config object from yaml file"""
        return _load_config_from_yaml(path, config_class=cls, **kwargs)

    @classmethod
    def from_hf(cls, repo_id, path, **kwargs):
        """Load config object from huggingface hub.

        Example:

        .. code-block:: python

            config = Config.from_hf("zeahub/configs", "config_camus.yaml", repo_type="dataset")

        Args:
            repo_id (str): huggingface hub repo id.
                For example: "zeahub/configs"
            path (str): path to the config file in the repo.
                For example: "train_config.yaml"
            **kwargs: additional arguments to pass to the `hf_hub_download`
                function. For example, use repo_type="dataset" to download
                from a dataset repo, or revision="main" to download from
                a specific branch.

        Returns:
            Config: config object.
        """
        local_path = hf_hub_download(repo_id, path, **kwargs)
        return _load_config_from_yaml(local_path, config_class=cls)

    def to_tensor(self):
        """Convert the attributes in the object to keras tensors"""
        return object_to_tensor(self)


def check_config(config: Union[dict, Config], verbose: bool = False):
    """Check a config given dictionary"""

    def _try_validate_config(config):
        try:
            config = config_schema.validate(config)
            return config
        except Exception as e:
            log.error(f"Config is not valid: {e}")
            raise e

    assert type(config) in [
        dict,
        Config,
    ], f"Config must be a dictionary or Config object, not {type(config)}"
    if isinstance(config, Config):
        config = config.serialize()
        config = _try_validate_config(config)
        config = Config(config)
        config.freeze()  # freeze because schema will add all defaults
    else:
        config = _try_validate_config(config)
    if verbose:
        log.success("Config is correct")
    return config


def _load_config_from_yaml(path, config_class=Config, loader=yaml.FullLoader):
    """Load config object from yaml file

    Args:
        path (str): path to yaml file.
        loader (yaml.Loader, optional): yaml loader. Defaults to yaml.FullLoader.
            for custom objects, you might want to use yaml.UnsafeLoader.
        config_class (type, optional): Config class to instantiate. Defaults to Config.

    Returns:
        Config: config object.
    """
    with open(Path(path), "r", encoding="utf-8") as file:
        dictionary = yaml.load(file, Loader=loader)
    if dictionary:
        return config_class(dictionary)
    else:
        return config_class()


def _path_to_str(path):
    """Convert a Path object to a string."""
    if hasattr(path, "as_posix"):
        # If path is a Path object, convert to string
        path = path.as_posix()
    return path



"""General utility functions."""

import collections.abc
import datetime
import functools
import hashlib
import inspect
import platform
import time
from functools import wraps
from pathlib import Path
from statistics import mean, median, stdev

import numpy as np
import yaml
from keras import ops
from PIL import Image

from zea import log


def _assert_uint8_images(images: np.ndarray):
    """
    Asserts that the input images have the correct properties.

    Args:
        images (np.ndarray): The input images.

    Raises:
        AssertionError: If the dtype of images is not uint8.
        AssertionError: If the shape of images is not (n_frames, height, width, channels)
            or (n_frames, height, width) for grayscale images.
        AssertionError: If images have anything other than 1 (grayscale),
            3 (rgb) or 4 (rgba) channels.
    """
    assert images.dtype == np.uint8, f"dtype of images should be uint8, got {images.dtype}"

    assert images.ndim in (3, 4), (
        "images must have shape (n_frames, height, width, channels),"
        f" or (n_frames, height, width) for grayscale images. Got {images.shape}"
    )

    if images.ndim == 4:
        assert images.shape[-1] in (1, 3, 4), (
            "Grayscale images must have 1 channel, "
            "RGB images must have 3 channels, and RGBA images must have 4 channels. "
            f"Got shape: {images.shape}, channels: {images.shape[-1]}"
        )


def translate(array, range_from=None, range_to=(0, 255)):
    """Map values in array from one range to other.

    Args:
        array (ndarray): input array.
        range_from (Tuple, optional): lower and upper bound of original array.
            Defaults to min and max of array.
        range_to (Tuple, optional): lower and upper bound to which array should be mapped.
            Defaults to (0, 255).

    Returns:
        (ndarray): translated array
    """
    if range_from is None:
        left_min, left_max = ops.min(array), ops.max(array)
    else:
        left_min, left_max = range_from
    right_min, right_max = range_to

    # Convert the left range into a 0-1 range (float)
    value_scaled = (array - left_min) / (left_max - left_min)

    # Convert the 0-1 range into a value in the right range.
    return right_min + (value_scaled * (right_max - right_min))


def map_negative_indices(indices: list, length: int):
    """Maps negative indices for array indexing to positive indices.
    Example:
        >>> from zea.utils import map_negative_indices
        >>> map_negative_indices([-1, -2], 5)
        [4, 3]
    """
    return [i if i >= 0 else length + i for i in indices]


def find_key(dictionary, contains, case_sensitive=False):
    """Find key in dictionary that contains partly the string `contains`

    Args:
        dictionary (dict): Dictionary to find key in.
        contains (str): String which the key should .
        case_sensitive (bool, optional): Whether the search is case sensitive.
            Defaults to False.

    Returns:
        str: the key of the dictionary that contains the query string.

    Raises:
        TypeError: if not all keys are strings.
        KeyError: if no key is found containing the query string.
    """
    # Assert that all keys are strings
    if not all(isinstance(k, str) for k in dictionary.keys()):
        raise TypeError("All keys must be strings.")

    if case_sensitive:
        key = [k for k in dictionary.keys() if contains in k]
    else:
        key = [k for k in dictionary.keys() if contains in k.lower()]

    if len(key) == 0:
        raise KeyError(f"Key containing '{contains}' not found in dictionary.")

    return key[0]


def print_clear_line():
    """Clears line. Helpful when printing in a loop on the same line."""
    line_up = "\033[1A"
    line_clear = "\x1b[2K"
    print(line_up, end=line_clear)


def strtobool(val: str):
    """Convert a string representation of truth to True or False.

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    assert isinstance(val, str), f"Input value must be a string, not {type(val)}"
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif val in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError(f"invalid truth value {val}")


def grayscale_to_rgb(image):
    """Converts a grayscale image to an RGB image.

    Args:
        image (ndarray): Grayscale image. Must have shape (height, width).

    Returns:
        ndarray: RGB image.
    """
    assert image.ndim == 2, "Input image must be grayscale."
    # Stack the grayscale image into 3 channels (RGB)
    return np.stack([image] * 3, axis=-1)


def preprocess_for_saving(images):
    """Preprocesses images for saving to GIF or MP4.

    Args:
        images (ndarray, list[ndarray]): Images. Must have shape (n_frames, height, width, channels)
            or (n_frames, height, width).
    """
    images = np.array(images)
    _assert_uint8_images(images)

    # Remove channel axis if it is 1 (grayscale image)
    if images.ndim == 4 and images.shape[-1] == 1:
        images = np.squeeze(images, axis=-1)

    # convert grayscale images to RGB
    if images.ndim == 3:
        images = [grayscale_to_rgb(image) for image in images]
        images = np.array(images)

    return images


def save_to_gif(images, filename, fps=20, shared_color_palette=False):
    """Saves a sequence of images to a GIF file.

    Args:
        images (list or np.ndarray): List or array of images. Must have shape
            (n_frames, height, width, channels) or (n_frames, height, width).
            If channel axis is not present, or is 1, grayscale image is assumed,
            which is then converted to RGB. Images should be uint8.
        filename (str or Path): Filename to which data should be written.
        fps (int): Frames per second of rendered format.
        shared_color_palette (bool, optional): If True, creates a global
            color palette across all frames, ensuring consistent colors
            throughout the GIF. Defaults to False, which is default behavior
            of PIL.Image.save. Note: True can cause slow saving for longer sequences.

    """
    images = preprocess_for_saving(images)

    if fps > 50:
        log.warning(f"Cannot set fps ({fps}) > 50. Setting it automatically to 50.")
        fps = 50

    duration = 1 / (fps) * 1000  # milliseconds per frame

    pillow_imgs = [Image.fromarray(img) for img in images]

    if shared_color_palette:
        # Apply the same palette to all frames without dithering for consistent color mapping
        # Convert all images to RGB and combine their colors for palette generation
        all_colors = np.vstack([np.array(img.convert("RGB")).reshape(-1, 3) for img in pillow_imgs])
        combined_image = Image.fromarray(all_colors.reshape(-1, 1, 3))

        # Generate palette from all frames
        global_palette = combined_image.quantize(
            colors=256,
            method=Image.MEDIANCUT,
            kmeans=1,
        )

        # Apply the same palette to all frames without dithering
        pillow_imgs = [
            img.convert("RGB").quantize(
                palette=global_palette,
                dither=Image.NONE,
            )
            for img in pillow_imgs
        ]

    pillow_img, *pillow_imgs = pillow_imgs

    pillow_img.save(
        fp=filename,
        format="GIF",
        append_images=pillow_imgs,
        save_all=True,
        loop=0,
        duration=duration,
        interlace=False,
        optimize=False,
    )
    log.success(f"Succesfully saved GIF to -> {log.yellow(filename)}")


def save_to_mp4(images, filename, fps=20):
    """Saves a sequence of images to an MP4 file.

    Args:
        images (list or np.ndarray): List or array of images. Must have shape
            (n_frames, height, width, channels) or (n_frames, height, width).
            If channel axis is not present, or is 1, grayscale image is assumed,
            which is then converted to RGB. Images should be uint8.
        filename (str or Path): Filename to which data should be written.
        fps (int): Frames per second of rendered format.

    Returns:
        str: Success message.

    """
    images = preprocess_for_saving(images)

    filename = str(filename)

    parent_dir = Path(filename).parent
    if not parent_dir.exists():
        raise FileNotFoundError(f"Directory '{parent_dir}' does not exist.")

    try:
        import cv2
    except ImportError as exc:
        raise ImportError(
            "OpenCV is required to save MP4 files. "
            "Please install it with 'pip install opencv-python' or "
            "'pip install opencv-python-headless'."
        ) from exc

    fourcc = cv2.VideoWriter_fourcc(*"mp4v")
    _, height, width, _ = images.shape
    video_writer = cv2.VideoWriter(filename, fourcc, fps, (width, height))

    for image in images:
        video_writer.write(image)

    video_writer.release()
    return log.success(f"Successfully saved MP4 to -> {filename}")


def update_dictionary(dict1: dict, dict2: dict, keep_none: bool = False) -> dict:
    """Updates dict1 with values dict2

    Args:
        dict1 (dict): base dictionary
        dict2 (dict): update dictionary
        keep_none (bool, optional): whether to keep keys
            with None values in dict2. Defaults to False.

    Returns:
        dict: updated dictionary
    """
    if not keep_none:
        dict2 = {k: v for k, v in dict2.items() if v is not None}
    # dict merging python > 3.9: default_scan_params | config_scan_params
    dict_out = {**dict1, **dict2}
    return dict_out


def get_date_string(string: str = None):
    """Generate a date string for current time, according to format specified by
    `string`. Refer to the documentation of the datetime module for more information
    on the formatting options.

    If no string is specified, the default format is used: "%Y_%m_%d_%H%M%S".
    """
    if string is not None and not isinstance(string, str):
        raise TypeError("Input must be a string.")

    # Get the current time
    now = datetime.datetime.now()

    # If no string is specified, use the default format
    if string is None:
        string = "%Y_%m_%d_%H%M%S"

    # Generate the date string
    date_str = now.strftime(string)

    return date_str


def date_string_to_readable(date_string: str, include_time: bool = False):
    """Converts a date string to a more readable format.

    Args:
        date_string (str): The input date string.
        include_time (bool, optional): Whether to include the time in the output.
            Defaults to False.

    Returns:
        str: The date string in a more readable format.
    """
    date = datetime.datetime.strptime(date_string, "%Y_%m_%d_%H%M%S")
    if include_time:
        return date.strftime("%B %d, %Y %I:%M %p")
    else:
        return date.strftime("%B %d, %Y")


def find_first_nonzero_index(arr, axis, invalid_val=-1):
    """
    Find the index of the first non-zero element along a specified axis in a NumPy array.

    Args:
        arr (numpy.ndarray): The input array to search for the first non-zero element.
        axis (int): The axis along which to perform the search.
        invalid_val (int, optional): The value to assign to elements where no
            non-zero values are found along the axis.

    Returns:
        numpy.ndarray: An array of indices where the first non-zero element
            occurs along the specified axis. Elements with no non-zero values along
            the axis are replaced with the 'invalid_val'.

    """
    nonzero_mask = arr != 0
    return np.where(nonzero_mask.any(axis=axis), nonzero_mask.argmax(axis=axis), invalid_val)


def first_not_none_item(arr):
    """
    Finds and returns the first non-None item in the given array.

    Args:
        arr (list): The input array.

    Returns:
        The first non-None item found in the array, or None if no such item exists.
    """
    non_none_items = [item for item in arr if item is not None]
    return non_none_items[0] if non_none_items else None


def deprecated(replacement=None):
    """Decorator to mark a function, method, or attribute as deprecated.

    Args:
        replacement (str, optional): The name of the replacement function, method, or attribute.

    Returns:
        callable: The decorated function, method, or property.

    Raises:
        DeprecationWarning: A warning is issued when the deprecated item is called or accessed.

    Example:
        >>> from zea.utils import deprecated
        >>> class MyClass:
        ...     @deprecated(replacement="new_method")
        ...     def old_method(self):
        ...         print("This is the old method.")
        ...
        ...     @deprecated(replacement="new_attribute")
        ...     def __init__(self):
        ...         self._old_attribute = "Old value"
        ...
        ...     @deprecated(replacement="new_property")
        ...     @property
        ...     def old_property(self):
        ...         return self._old_attribute

        >>> # Using the deprecated method
        >>> obj = MyClass()
        >>> obj.old_method()
        This is the old method.
        >>> # Accessing the deprecated attribute
        >>> print(obj.old_property)
        Old value
        >>> # Setting value to the deprecated attribute
        >>> obj.old_property = "New value"
    """

    def decorator(item):
        if callable(item):
            # If it's a function or method
            @functools.wraps(item)
            def wrapper(*args, **kwargs):
                if replacement:
                    log.deprecated(
                        f"Call to deprecated {item.__name__}. Use {replacement} instead."
                    )
                else:
                    log.deprecated(f"Call to deprecated {item.__name__}.")
                return item(*args, **kwargs)

            return wrapper
        elif isinstance(item, property):
            # If it's a property of a class
            def getter(self):
                if replacement:
                    log.deprecated(
                        f"Access to deprecated attribute {item.fget.__name__}, "
                        f"use {replacement} instead."
                    )
                else:
                    log.deprecated(f"Access to deprecated attribute {item.fget.__name__}.")
                return item.fget(self)

            def setter(self, value):
                if replacement:
                    log.deprecated(
                        f"Setting value to deprecated attribute {item.fget.__name__}, "
                        f"use {replacement} instead."
                    )
                else:
                    log.deprecated(f"Setting value to deprecated attribute {item.fget.__name__}.")
                item.fset(self, value)

            def deleter(self):
                if replacement:
                    log.deprecated(
                        f"Deleting deprecated attribute {item.fget.__name__}, "
                        f"use {replacement} instead."
                    )
                else:
                    log.deprecated(f"Deleting deprecated attribute {item.fget.__name__}.")
                item.fdel(self)

            return property(getter, setter, deleter)

        else:
            raise TypeError("Decorator can only be applied to functions, methods, or properties.")

    return decorator


def calculate_file_hash(file_path, omit_line_str=None):
    """Calculates the hash of a file.

    Args:
        file_path (str): Path to file.
        omit_line_str (str, optional): If this string is found in a line, the line will
            be omitted when calculating the hash. This is useful for example
            when the file contains the hash itself.

    Returns:
        str: The hash of the file.

    """
    hash_object = hashlib.sha256()
    with open(file_path, "rb") as f:
        for line in f:
            if omit_line_str is not None and omit_line_str in str(line):
                continue
            hash_object.update(line)
    return hash_object.hexdigest()


def check_architecture():
    """Checks the architecture of the system."""
    return platform.uname()[-1]


def get_function_args(func):
    """Get the names of the arguments of a function."""
    sig = inspect.signature(func)
    return tuple(sig.parameters)


def fn_requires_argument(fn, arg_name):
    """Returns True if the function requires the argument 'arg_name'."""
    params = get_function_args(fn)
    return arg_name in params


def find_methods_with_return_type(cls, return_type_hint):
    """
    Find all methods in a class that have the specified return type hint.

    Args:
        cls: The class to inspect.
        return_type_hint: The return type hint to match (as a string).

    Returns:
        A list of method names that match the return type hint.
    """
    matching_methods = []
    for name, member in inspect.getmembers(cls, predicate=inspect.isfunction):
        annotations = getattr(member, "__annotations__", {})
        if annotations.get("return") == return_type_hint:
            matching_methods.append(name)
    return matching_methods


def keep_trying(fn, args=None, required_set=None):
    """Keep trying to run a function until it succeeds.

    Args:
        fn (callable): Function to run.
        args (dict, optional): Arguments to pass to function.
        required_set (set, optional): Set of required outputs.
            If output is not in required_set, function will be rerun.

    Returns:
        Any: The output of the function if successful.

    """
    while True:
        try:
            out = fn(**args) if args is not None else fn()
            if required_set is not None:
                assert out is not None
                assert out in required_set, f"Output {out} not in {required_set}"
            return out
        except Exception as e:
            print(e)


def reduce_to_signature(func, kwargs):
    """Reduce the kwargs to the signature of the function."""
    # Retrieve the argument names of the function
    sig = inspect.signature(func)

    # Filter out the arguments that are not part of the function
    reduced_params = {key: kwargs[key] for key in sig.parameters if key in kwargs}

    return reduced_params


def safe_initialize_class(cls, **kwargs):
    """Safely initialize a class by removing any invalid arguments."""

    # Filter out the arguments that are not part of the Scan class
    reduced_params = reduce_to_signature(cls.__init__, kwargs)

    return cls(**reduced_params)


def deep_compare(obj1, obj2):
    """Recursively compare two objects for equality."""
    # Only recurse into dicts
    if isinstance(obj1, dict) and isinstance(obj2, dict):
        if obj1.keys() != obj2.keys():
            return False
        return all(deep_compare(obj1[k], obj2[k]) for k in obj1)

    # If not dict, but both are iterable (excluding strings/bytes), compare items
    if (
        isinstance(obj1, collections.abc.Iterable)
        and isinstance(obj2, collections.abc.Iterable)
        and not isinstance(obj1, (str, bytes))
        and not isinstance(obj2, (str, bytes))
    ):
        return all(deep_compare(a, b) for a, b in zip(obj1, obj2))

    # Fallback to direct comparison (also handles int, float, str, etc.)
    return obj1 == obj2


class FunctionTimer:
    """
    A decorator class for timing the execution of functions.

    Example:
        >>> from zea.utils import FunctionTimer
        >>> timer = FunctionTimer()
        >>> my_function = lambda: sum(range(10))
        >>> my_function = timer(my_function)
        >>> _ = my_function()
        >>> print(timer.get_stats("my_function"))
    """

    def __init__(self):
        self.timings = {}
        self.last_append = 0

    def __call__(self, func, name=None):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start_time = time.perf_counter()
            result = func(*args, **kwargs)
            end_time = time.perf_counter()
            elapsed_time = end_time - start_time

            # Store the timing result
            _name = name if name is not None else func.__name__
            if _name not in self.timings:
                self.timings[_name] = []
            self.timings[_name].append(elapsed_time)

            return result

        return wrapper

    def get_stats(self, func_name, drop_first: bool | int = False):
        """Calculate statistics for the given function."""
        if func_name not in self.timings:
            raise ValueError(f"No timings recorded for function '{func_name}'.")

        if isinstance(drop_first, bool):
            idx = 1 if drop_first else 0
        elif isinstance(drop_first, int):
            idx = drop_first
        else:
            raise ValueError("drop_first must be a boolean or an integer.")

        times = self.timings[func_name][idx:]
        return {
            "mean": mean(times),
            "median": median(times),
            "std_dev": stdev(times) if len(times) > 1 else 0,
            "min": min(times),
            "max": max(times),
            "count": len(times),
        }

    def export_to_yaml(self, filename):
        """Export the timing data to a YAML file."""
        with open(filename, "w", encoding="utf-8") as f:
            yaml.dump(self.timings, f, default_flow_style=False)
        print(f"Timing data exported to '{filename}'.")

    def append_to_yaml(self, filename, func_name):
        """Append the timing data to a YAML file."""
        cropped_timings = self.timings[func_name][self.last_append :]

        with open(filename, "a", encoding="utf-8") as f:
            yaml.dump(cropped_timings, f, default_flow_style=False)

        self.last_append = len(self.timings[func_name])



"""
zea setup
===========

This module provides setup functions for the zea package, grouping together
commonly used initialization routines for convenience.

Overview
--------

The main entry point is the :func:`setup` function, which performs several key
initialization steps for you:

- Loads and validates the configuration file (YAML) via
    :func:`zea.internal.setup_zea.setup_config`.
- Sets up user data paths using :func:`zea.datapaths.set_data_paths`.
- Initializes the device (GPU/CPU) according to the configuration using
    :func:`zea.internal.device.init_device`.
- Optionally creates a new user and prompts for datapath information via
    :func:`zea.datapaths.create_new_user`.

By calling :func:`setup`, you can prepare your zea environment in a single step,
ensuring that configuration, data paths, and device setup are all handled for you.

.. code-block:: python


    # Basic usage: loads config, sets paths, initializes device
    config = setup_zea.setup(config_path="my_config.yaml")

    # With user creation prompt
    config = setup_zea.setup(config_path="my_config.yaml", create_user=True)

Function Details
----------------

.. autofunction:: setup
    :noindex:

    Calls:
        - :func:`zea.internal.setup_zea.setup_config`
        - :func:`zea.datapaths.set_data_paths`
        - :func:`zea.internal.device.init_device`
        - :func:`zea.datapaths.create_new_user` (if ``create_user=True``)

.. autofunction:: setup_config
    :noindex:

    Loads and validates the configuration file for zea. Supports interactive
    selection if no path is provided.

"""

import copy
import importlib
import sys
from pathlib import Path
from typing import Union

import keras
import yaml

from zea import Config, log
from zea.config import check_config
from zea.data.preset_utils import HF_PREFIX, _hf_parse_path
from zea.datapaths import create_new_user, set_data_paths
from zea.internal.device import init_device
from zea.internal.git_info import get_git_summary
from zea.internal.viewer import filename_from_window_dialog


def reload_module(name):
    """Reloads module. This is useful when changing the backend.
    Taken from `keras.config.set_backend`"""

    # Clear module cache.
    loaded_modules = [key for key in sys.modules if key.startswith(name)]
    for key in loaded_modules:
        del sys.modules[key]

    __class__ = importlib.import_module(name).__class__

    # Finally: refresh all imported submodules.
    globs = copy.copy(globals())
    for key, value in globs.items():
        if value.__class__ == __class__:
            if str(value).startswith(f"<module '{name}."):
                module_name = str(value)
                module_name = module_name[module_name.find("'") + 1 :]
                module_name = module_name[: module_name.find("'")]
                globals()[key] = importlib.import_module(module_name)


def set_backend(backend: str):
    """Set compute backend

    Note: Make sure to reimport any module you are using that uses keras
    directly (has import keras or derivative at the top of the file).
    This can be done with the importlib module.

    """
    # set keras backend
    if keras.config.backend() != backend:
        keras.config.set_backend(backend)
        reload_module("zea")


def setup(
    config_path: str = None,
    user_config: Union[str, dict] = None,
    verbose: bool = True,
    disable_config_check: bool = False,
    loader=yaml.FullLoader,
    create_user: bool = False,
):
    """General setup function for zea. Loads config, sets data paths and
    initializes gpu if available. Will return config object.

    Args:
        config_path (str, optional): file path to config yaml.
            Defaults to None, in which case a window dialog will pop up.
        user_config (str or dict, optional): path that points to yaml file with user info.
            Alternively dictionary with user info. Defaults to None.
        verbose (bool, optional): print config file path and git summary. Defaults to True.
        disable_config_check (bool, optional): whether to check for zea config validity.
            Defaults to False. Can be set to True if you are using some other config that
            does not have to adhere to zea config standards.
        loader (yaml.Loader, optional): yaml loader. Defaults to yaml.FullLoader.
        create_user (bool, optional): whether to create a new user. Defaults to False.
            If True, it will prompt the user to enter their datapaths.
    Returns:
        config (dict): config object / dict.
    """

    # Load config
    config = setup_config(
        config_path,
        verbose=verbose,
        disable_config_check=disable_config_check,
        loader=loader,
    )

    # Prompt user to enter datapath information
    if create_user:
        create_new_user(user_config, local=config.data.local)

    # Set data paths
    config.data.user = set_data_paths(user_config, local=config.data.local)

    # Init GPU / CPU according to config
    config.device = init_device(
        device=config.device,
        backend="auto",
        hide_devices=config.hide_devices,
    )

    return config


def setup_config(
    config_path: str = None,
    verbose: bool = True,
    disable_config_check: bool = False,
    loader=yaml.FullLoader,
    repo_type: str = "dataset",
):
    """Setup function for config. Retrieves config file and checks for validity.

    Args:
        config_path (str, optional): file path to config yaml. Defaults to None.
            if None, argparser is checked. If that is None as well, the window
            ui will pop up for choosing the config file manually. Besides a local
            path it can also be a Hugging Face repository path, in the form
            hf://<repo_id>/<path_to_config>. For example:
            `hf://username/zeahub/configs/config_camus.yaml`.
        verbose (bool, optional): print config file path and git summary. Defaults to True.
        disable_config_check (bool, optional): whether to check for zea config validity.
            Defaults to False. Can be set to True if you are using some other config that
            does not have to adhere to zea config standards.
        loader (yaml.Loader, optional): yaml loader. Defaults to yaml.FullLoader.
            for custom objects, you might want to use yaml.UnsafeLoader.
        repo_type (str, optional): type of the Hugging Face repository.
            Defaults to "dataset". Only used when `config_path` starts with
            `HF_PREFIX`. Can be "model", "dataset", or "space".
    Returns:
        config (dict): config object / dict.

    """
    if config_path is None:
        # if no argument is provided resort to UI window
        filetype = "yaml"
        try:
            config_path = filename_from_window_dialog(
                f"Choose .{filetype} file",
                filetypes=((filetype, "*." + filetype),),
                initialdir="./configs",
            )
        except Exception as e:
            raise ValueError(
                "Please specify the path to a config file through --config flag "
                "i.e. `zea --config <path-to-config.yaml>` if GUI is not working "
                "(usually on headless servers)."
            ) from e

    if str(config_path).startswith(HF_PREFIX):
        repo_id, path = _hf_parse_path(config_path)
        config = Config.from_hf(
            repo_id=repo_id,
            path=path,
            repo_type=repo_type,
        )
    else:
        config = Config.from_yaml(Path(config_path), loader=loader)

    if verbose:
        log.info(f"Using config file: {log.yellow(config_path)}")

    config["git"] = get_git_summary(verbose=verbose)

    if not disable_config_check:
        config = check_config(config)

    return config



"""Utility functions for zea tensorflow modules."""

import numpy as np
import tensorflow as tf


class DotDict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def tf_snapshot(obj) -> dict:
    """Returns a snapshot of the object parameters as a dictionary of tensors.

    Returns:
        dict: The scan parameters as a dictionary of tensors.
    """
    EXCEPTIONS = ["angles", "_angles"]

    snapshot = DotDict()
    for key in dir(obj):
        if key[0] != "_" and key not in EXCEPTIONS:
            value = getattr(obj, key)
            if isinstance(value, (np.ndarray, int, float, list)):
                # if data is of double precision, convert to float32
                if isinstance(value, np.ndarray) and value.dtype == np.float64:
                    dtype = tf.float32
                else:
                    dtype = None

                snapshot[key] = tf.convert_to_tensor(value, dtype=dtype)
    return snapshot



"""This module contains functionality for converting between hdf5, mat and
dictionary.

"""

import argparse
from pathlib import Path

import h5py
import numpy as np
import scipy.io as sio

from zea.data.file import recursively_load_dict_contents_from_group
from zea.internal.viewer import filename_from_window_dialog


def load_mat(filename):
    """
    This function should be called instead of direct scipy.io.loadmat
    as it cures the problem of not properly recovering python dictionaries
    from mat files. It calls the function check keys to cure all entries
    which are still mat-objects
    """

    def _check_vars(d):
        """
        Checks if entries in dictionary are mat-objects. If yes
        todict is called to change them to nested dictionaries
        """
        for key in d:
            if isinstance(d[key], sio.matlab.mio5_params.mat_struct):
                d[key] = _todict(d[key])
            elif isinstance(d[key], np.ndarray):
                d[key] = _toarray(d[key])
        return d

    def _todict(matobj):
        """
        A recursive function which constructs from matobjects nested dictionaries
        """
        d = {}
        for strg in matobj._fieldnames:
            elem = matobj.__dict__[strg]
            if isinstance(elem, sio.matlab.mio5_params.mat_struct):
                d[strg] = _todict(elem)
            elif isinstance(elem, np.ndarray):
                d[strg] = _toarray(elem)
            else:
                d[strg] = elem
        return d

    def _toarray(ndarray):
        """
        A recursive function which constructs ndarray from cellarrays
        (which are loaded as numpy ndarrays), recursing into the elements
        if they contain matobjects.
        """
        if ndarray.dtype != "float64":
            elem_list = []
            for sub_elem in ndarray:
                if isinstance(sub_elem, sio.matlab.mio5_params.mat_struct):
                    elem_list.append(_todict(sub_elem))
                elif isinstance(sub_elem, np.ndarray):
                    elem_list.append(_toarray(sub_elem))
                else:
                    elem_list.append(sub_elem)
            return np.array(elem_list)
        else:
            return ndarray

    data = sio.loadmat(filename, struct_as_record=False, squeeze_me=True)
    return _check_vars(data)


def save_dict_to_file(filename, dic):
    """Save dict to .mat or .hdf5"""

    filetype = Path(filename).suffix
    assert filetype in [".mat", ".hdf5"]

    if filetype == ".hdf5":
        with h5py.File(filename, "w") as h5file:
            recursively_save_dict_contents_to_group(h5file, "/", dic)
    elif filetype == ".mat":
        sio.savemat(filename, dic)


def recursively_save_dict_contents_to_group(h5file, path, dic):
    """Save dict contents to group"""
    for key, item in dic.items():
        if isinstance(item, dict):
            recursively_save_dict_contents_to_group(h5file, path + key + "/", item)
        else:
            h5file[path + key] = item


def load_dict_from_file(filename, squeeze=True):
    """dict from file"""
    filetype = Path(filename).suffix
    assert filetype in [".mat", ".hdf5"]

    v_7_3 = False
    if filetype == ".mat":
        try:
            return load_mat(filename)
        except Exception:
            v_7_3 = True

    if (filetype == ".hdf5") or (v_7_3 is True):
        with h5py.File(filename, "r", locking=False) as h5file:
            return recursively_load_dict_contents_from_group(h5file, "/", squeeze)


def strip_matfile_keys(dic):
    """Strip some unecessary .mat keys"""
    new_dic = {}
    for key, val in dic.items():
        if key not in ["__globals__", "__header__", "__version__"]:
            new_dic[key] = np.squeeze(val)
    return new_dic


def get_args():
    """Command line argument parser"""
    parser = argparse.ArgumentParser()
    parser.add_argument("--file", default=None, help="hdf5 file or mat file")
    args = parser.parse_args()
    return args


if __name__ == "__main__":
    args = get_args()
    data = {}
    if args.file is None:
        file = filename_from_window_dialog(
            "Choose .mat or .hdf5 file",
            filetypes=(("mat or hdf5", "*.mat *.hdf5 *.h5"),),
        )
    else:
        file = Path(args.file)

    dic = load_dict_from_file(file)
    if file.suffix == ".mat":
        dic = strip_matfile_keys(dic)
        save_dict_to_file(file.with_suffix(".hdf5"), dic)

    elif file.suffix in [".h5", ".hdf5"]:
        save_dict_to_file(file.with_suffix(".mat"), dic)

    print(f"Succesfully converted {file}")
